#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurusdevicetree.py:
"""

# @todo: This module is not being used anywhere in Taurus and depends on
# non-standard and non-provided modules. It is also quite specific and
# tango-centric, so it should be removed from Taurus or merged with
# Taurusdbtree

# ,"SearchEdit"] #"TaurusTreeNode"]
__all__ = ["TaurusDevTree", "TaurusSearchTree", "TaurusDevTreeOptions"]

import time
import os
import traceback
from functools import partial
import PyTango  # to change!!

# @todo: icons_dev_tree is not an included or standard module.
#        Is anybody using it? If not, the following lines should be removed and
#        the TaurusDevTree.setStateIcon method should be cleaned
try:
    import icons_dev_tree
except:
    icons_dev_tree = None

from taurus.external.qt import Qt

import taurus.core
from taurus.core.util.colors import DEVICE_STATE_PALETTE, ATTRIBUTE_QUALITY_PALETTE
from taurus.core.util.containers import CaselessDict
from taurus.core.tango.search import *  # @TODO: Avoid implicit imports
from taurus.qt.qtcore.util.emitter import SingletonWorker
from taurus.qt.qtcore.mimetypes import *  # @TODO: Avoid implicit imports
from taurus.qt.qtcore.util import properties
from taurus.qt.qtcore.util.properties import djoin
from taurus.qt.qtgui.base import TaurusBaseComponent, TaurusBaseWidget
from taurus.qt.qtgui.container import TaurusWidget

TREE_ITEM_MIME_TYPE = 'application/x-qabstractitemmodeldatalist'

###############################################################################


class TaurusTreeNodeContainer(object):
    """
    Interface that provides Node-focused methods to TaurusDevTree
    This methods should be moved to TaurusDevTreeNode class when it will be able to retrieve currentItem from Tree.
    """
    _icon_map = {}  # A dictionary like {device_regexp:pixmap_url}

    addAttrSelected = Qt.pyqtSignal('QStringList')
    removeAttrSelected = Qt.pyqtSignal('QStringList')

    def __init__(self):
        raise Exception(
            'This class is just an interface, do not instantiate it!')

    @classmethod
    def setIconMap(klass, filters):
        """A dictionary like {device_regexp:pixmap_url}"""
        klass._icon_map = filters

    @classmethod
    def getIconMap(klass):
        return klass._icon_map

    def createItem(self, parent, value, text=None):
        self.debug('createItem(%s,%s)' % (value, text))
        USE_TREE_NODE = False
        if USE_TREE_NODE:
            item = TaurusTreeNode(parent)
        else:
            item = Qt.QTreeWidgetItem(parent)
        if text is None:
            text = value
        item.isAttribute = False
        item.DeviceName = ''
        item.draggable = ''
        item.setText(0, Qt.QApplication.translate(
            '', text, None, Qt.QApplication.UnicodeUTF8))
        self.setNodeParent(item, parent)
        item.adminNode = None
        if not item.parentNode or '/' in text:
            f = item.font(0)
            if not item.parentNode:
                f.setBold(True)
            if '/' in text:
                f.setItalic(True)
            item.setFont(0, f)
        item.parentTree = self  # hook used to call external methods with item as single argument
        self.item_index[value.strip().split()[0]] = item
        try:
            icon = self.getNodeIcon(item)
            if icon:
                item.setIcon(0, icon)
        except:
            pass
        self.item_list.add(item)
        return item

    ###########################################################################
    # Item members methods

    def setNodeParent(self, node, parent):
        """ Used to know which parent attributes must be expanded if found """
        node.parentNode = parent if isinstance(
            parent, Qt.QTreeWidgetItem) else None

    def setNodeAdmin(self, node, admin):
        """ Used to assign a controller to its controlled devices in the tree """
        node.adminNode = admin.getNodeText(admin) if isinstance(
            admin, Qt.QTreeWidgetItem) else None

    def getNodeAdmin(self, node):
        return node.adminNode(node) if isCallable(node.adminNode) else node.adminNode

    def getNodeText(self, node=None, full=False):
        """ Get the text of the node as shown in the tree, @full allows to get the first word or the whole text"""
        if node is None:
            node = self.currentItem()
        if hasattr(node, 'text'):
            txt = str(node.text(0)).strip()
            if not full:
                return txt.split()[0]
            return txt
        else:
            return ''

    def getNodeDeviceName(self, node=None):
        if node is None:
            node = self.currentItem()
        return str(getattr(node, 'DeviceName', '')) or self.getNodeText(node)

    def getNodeParentName(self, node=None):
        if node is None:
            node = self.currentItem()
        return self.getNodeText(node.parentNode)

    def getNodePath(self, node=None):
        """ Returns all parent nodes prior to current """
        if node is None:
            node = self.currentItem()
        p, path, names = node.parentNode, [], []
        while p is not None:
            path.insert(0, p)
            names.insert(0, self.getNodeDeviceName(p))
            p = p.parentNode
        return path

    def getNodeAlias(self, node=None):
        if node is None:
            node = self.currentItem()
        alias = getattr(node, 'AttributeAlias', '')
        return (alias or self.getNodeText(node))

    def getNodeIcon(self, node=None):
        #self.debug('TaurusDevTree.getNodeIcon(node) not implemented, overrided in subclasses')

        #self,url = node.parentTree,''
        if node is None:
            node = self.getNode()
        try:
            name, url = self.getNodeText(node), ''
            for k, v in self.getIconMap().items():
                if re.match(k.lower(), name.lower()):
                    url = v
            if not url:
                for k, v in self.getIconMap().items():
                    if k.lower() in name.lower():
                        url = v
            # if name.count('/')==2:
                # if any(a.startswith(name+'/') for a in getArchivedAttributes()):
                    #url = wdir('image/icons/clock.png')
                # else:
                    #url = wdir('image/equips/icon-%s.gif'%name.split('/')[2].split('-')[0].lower())
            # elif name.count('/')==3:
                #url = filterAttributes(name) or wdir('image/icons/closetab.png')
            # else:
                #url = wdir('image/equips/icon-%s.gif'%name.lower())
        except:
            self.warning(traceback.format_exc())
        if not url or not os.path.isfile(url):
            return None
        else:
            return Qt.QIcon(url)

    def getNodeDraggable(self, node=None):
        """ This method will return True only if the selected node belongs to a numeric Tango attribute """
        numtypes = [PyTango.DevDouble, PyTango.DevFloat, PyTango.DevLong, PyTango.DevLong64,
                    PyTango.DevULong, PyTango.DevShort, PyTango.DevUShort, PyTango.DevBoolean]
        if node is None:
            node = self.currentItem()
        try:
            name = self.getNodeText(node).lower()
            drag = name
            if node.isAttribute and getattr(node, 'DeviceName', '') and '/' not in name:
                name = node.DeviceName + '/' + name
            if name.count('/') == 2:  # A Device Name
                drag = name  # +'/state' #False
            elif name.count('/') == 3:  # An Attribute Name
                #dtype = PyTango.AttributeProxy(name).get_config().data_type
                #if dtype in numtypes: self.debug('The attribute %s is a Numeric Attribute'%(name))
                drag = getattr(node, 'draggable', '') or name
                # else: drag = False
            self.debug('Node(%s,%s,%s): drag: %s' %
                       (name, node.isAttribute, node.DeviceName, drag))
            return drag.split()[0]
        except:
            import traceback
            self.warning(traceback.format_exc())
            return False

    ###########################################################################
    # Context Menu Actions

    @staticmethod
    def setDefaultPanelClass(other):
        TaurusTreeNodeContainer._defaultClass = other

    @staticmethod
    def defaultPanelClass():
        if not hasattr(TaurusTreeNodeContainer, '_defaultClass'):
            from taurus.qt.qtgui.panel import TaurusDevicePanel
            TaurusTreeNodeContainer._defaultClass = TaurusDevicePanel
        obj = TaurusTreeNodeContainer._defaultClass
        return obj

    def showPanel(self):
        '''Display widget taurusDevicePanel'''
        device = self.getNodeText()
        nameclass = self.defaultPanelClass()()
        nameclass.setModel(device)
        nameclass.show()
        # nameclass.setSpectraAtkMode(True)
        # Dialog is used to make new floating panels persistent
        if isinstance(nameclass, TaurusWidget):
            PopupDialog(self, nameclass)

    def showProperties(self):
        '''Display widget TaurusPropTable'''
        import taurus.qt.qtgui.table
        device = self.getNodeText()
        nameclass = taurus.qt.qtgui.table.TaurusPropTable()
        nameclass.setTable(device)
        nameclass.show()
        # Dialog is used to make new floating panels persistent
        PopupDialog(self, nameclass)

    def addToPlot(self):
        """ This method will send a signal with the current selected node """
        items = self.getSelectedNodes()
        for item in items:
            attr = self.getNodeAlias(item)
            self.trace('In addToPlot(%s->%s)' % (item.text(0), attr))
            self.addAttrToPlot(attr)
        return

    def addAttrToPlot(self, attr):
        """ This method will send a signal with the given attr name, in a separate method to be called with a pre-filled list  """
        self.addAttrSelected.emit(Qt.QStringList([str(attr)]))

    def removeFromPlot(self):
        """ This method will send a signal with the current selected node """
        items = self.getSelectedNodes()
        for item in items:
            item = self.currentItem()
            attr = getattr(item, 'AttributeAlias',
                           '') or self.getNodeText(item)
            self.removeAttrFromPlot(attr)
        return

    def removeAttrFromPlot(self, attr):
        """ This method will send a signal with the given attr name, in a separate method to be called with a pre-filled list """
        self.removeAttrSelected.emit(Qt.QStringList([str(attr)]))


###############################################################################

class TaurusDevTree(TaurusTreeNodeContainer, Qt.QTreeWidget, TaurusBaseWidget):
    '''
    This widget displays a list of servers, devices or instances.
    To set a new Model use either setModel(filters), addModels(list), setFilters(...) or loadTree(filters)
    setModel and loadTree are equivalent; adding a new branch to the tree
    addModels merges the tree with new models
    setFilters clears previous models and adds new one
    '''
    __properties__ = (
        'ModelInConfig',
        'modifiableByUser',
        #'useParentModel',
        'Filters',
        'Source',
        'ShowAlias',
        'ShowColors',
        'ShowNotExported',
        'MaxDevices',
    )
    __slots__ = (
        "setTangoHost",
        #"setModel",
        #"setFilters",
        "addModels",
        "setModelCheck",
        "loadTree",  # Applies regexp filters to database
        "setTree",
        "findInTree",
        "setIcons",
        "expandAll",
        "expandNode",
        "collapseNode",
    )

    TRACE_ALL = False

    refreshTree = Qt.pyqtSignal()
    nodeFound = Qt.pyqtSignal()
    deviceSelected = Qt.pyqtSignal('QString')
    addAttrSelected = Qt.pyqtSignal('QStringList')
    removeAttrSelected = Qt.pyqtSignal('QStringList')

    def __init__(self, parent=None, designMode=False):
        name = "TaurusDevTree"
        self._useParentModel = True
        self._localModel = ''
        self.call__init__wo_kw(Qt.QTreeWidget, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)

        self.setObjectName(name)
        self._filters = []
        self.__attr_filter = None
        self.__expand = 1
        self.collapsing_search = True
        self.index = 0
        self._nameTopItem = ""

        # This is a list of regular expressions to exclude objects from
        # searches
        self.excludeFromSearch = []
        self.dictionary = {}
        self.item_index = CaselessDict()
        # NOTE: as several nodes may share the same name this list will be
        # different from item_index.values()!!!
        self.item_list = set()
        self.setSelectionMode(self.ExtendedSelection)

        self.ContextMenu = []
        self.ExpertMenu = []

        # The SingletonWorker Threads are used for expanding nodes and also for
        # loading a new tree; both objects are the same thread, but read from
        # different queues
        self.__loader = None
        self.__expander = None

        if not designMode:
            self.Loader
            self.Expander

        self.initConfig()

        # Signal
        self.itemclicked.connect(self.deviceClicked)
        self.nodeFound.connect(self.expandNode)
        self.setDragDropMode(Qt.QAbstractItemView.DragDrop)
        self.setModifiableByUser(True)
        self.setModelInConfig(False)  # We store Filters instead!
        self.setAcceptDrops(True)
        self.viewport().setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setSupportedMimeTypes([
            TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_DEV_MIME_TYPE, TAURUS_ATTR_MIME_TYPE,
            TAURUS_MODEL_MIME_TYPE, TREE_ITEM_MIME_TYPE, 'text/plain'])

        self.setTangoHost()
        self.defineStyle()

    @property
    def Loader(self):
        loader = self.__loader
        if loader is None:
            loader = SingletonWorker(parent=self, name='TreeLoader',
                                     cursor=True, start=True)
            self.__loader = loader
        return loader

    @property
    def Expander(self):
        expander = self.__expander
        if expander is None:
            expander = SingletonWorker(parent=self, name='NodeExpander',
                                       method=lambda node, expand: node.setExpanded(
                                           expand),
                                       cursor=True, start=True)
            self.__expander = expander
        return expander

    def getConfig(self, name):
        properties.get_property(self, name)

    def initConfig(self):
        """
        Initializing the attributes that will be kept persitent as Qt settings.
        e.g. for Filters property, the following attributes are created:

            - self.filters
            - self._filters
            - self.setFilters
            - self.getFilters
            - self.resetFilters

        """
        properties.set_property_methods(self, 'Models', 'QStringList', default='',
                                        #setter = self.setFilters,
                                        setter=self.addModels,  # Not trivial!; it avoids QSettings erasing default model
                                        #set_callback=lambda v,s=self:v and s.loadTree(v,clear=True),
                                        #reset_callback=lambda s=self:s.setFilters(''),
                                        qt=False, config=True
                                        )
        properties.set_property_methods(
            self, 'MaxDevices', 'int', default=150, qt=False, config=True)
        properties.set_property_methods(
            self, 'ShowAlias', 'bool', default=False, qt=False, config=True)
        properties.set_property_methods(
            self, 'ShowNotExported', 'bool', default=True, qt=False, config=True)
        properties.set_property_methods(
            self, 'ShowColors', 'bool', default=True, qt=False, config=True)
        # properties.set_property_methods(self,'Expand','int',default=0)

    @staticmethod
    def setDefaultAttrFilter(other):
        TaurusDevTree._defattrfilter = staticmethod(other)

    @staticmethod
    def defaultAttrFilter():
        if not hasattr(TaurusDevTree, '_defattrfilter'):
            TaurusDevTree._defattrfilter = None
        return TaurusDevTree._defattrfilter

    def setAttrFilter(self, other):
        self._attrfilter = other

    def getAttrFilter(self):
        if not isCallable(getattr(self, '_attrfilter', None)):
            self._attrfilter = None
        return self._attrfilter

    def matchAttrFilter(self, target):
        def printf(s):
            print(s)
        if self.getAttrFilter() and isCallable(self._attrfilter):
            return self._attrfilter(target, p=printf)
        elif TaurusDevTree.defaultAttrFilter() and isCallable(TaurusDevTree._defattrfilter):
            return TaurusDevTree._defattrfilter(target, p=printf)
        else:
            return True

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def sizeHint(self):
        return Qt.QTreeWidget.sizeHint(self)

    def minimumSizeHint(self):
        return Qt.QTreeWidget.minimumSizeHint(self)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.tree'
        ret['group'] = 'Taurus Views'
        ret['icon'] = "designer:listview.png"
        return ret

    def defineStyle(self):
        self.setWindowTitle('TaurusDevTree')
        self.setHeaderLabel(
            'Device Browser (right-click on any element to search/show options)')
        self.setGeometry(Qt.QRect(90, 60, 256, 192))
        self.actionFindInTree = Qt.QAction(self)
        self.actionFindInTree.setShortcut(Qt.QKeySequence.Find)
        self.actionFindInTree.triggered[()].connect(self.findDialog)
        #self.connect(self, Qt.SIGNAL("itemClicked"), self.clickedEvent)
        from taurus.qt.qtgui.table.qdictionary import QDictionaryEditor, QListEditor
        self.ExpertMenu.append(
            ('Edit Model Filters',
             lambda: QListEditor.main(
                 self._filters,
                 modal=True,
                 title='Edit Model Filters',
                 callback=lambda d: self.loadTree(d)
             )
                # lambda:self.loadTree(
                #str(Qt.QInputDialog.getText(None,'Set Tree Model','Enter a list of regexp separated by comma:',Qt.QLineEdit.Normal,','.join(str(f) for f in self._filters))[0])
                # or None)
             ))
        self.ExpertMenu.append(
            ('Edit Tree',
             lambda: QDictionaryEditor.main(
                 self.dictionary, modal=True, title='Edit Tree', callback=lambda d: self.setTree(d, clear=True))
             ))
        self.ExpertMenu.append(
            ('Expand All',
             lambda: self.expandAll()
             ))
        self.ExpertMenu.append(
            ('Collapse All',
             lambda: self.collapseNode(ALL=True)
             ))
        self.ExpertMenu.append(
            ('Save Config',
             lambda: self.saveConfigFile()
             ))
        if not getattr(self, 'DeviceMenu', None):
            self.DeviceMenu = {}
        self.DeviceMenu.update({
            'Show Properties': 'showProperties',
            'Refresh Tree': 'refreshTree',
        })
        if not getattr(self, 'AttributeMenu', None):
            self.AttributeMenu = []
        [self.AttributeMenu.append(a) for a in [
            ('Add to trends', 'addToPlot'),
            ('Remove from trends', 'removeFromPlot'),
        ] if a not in self.AttributeMenu]
        try:
            from PyTangoArchiving.widget.history import show_history
            self.debug('Adding show_history from archiving...')
            self.AttributeMenu.append(('Show History', show_history))
        except:
            pass

    def trace(self, msg):
        if self.TRACE_ALL or self.getLogLevel() in ('DEBUG', 40,):
            # @TODO: use the taurus logger instead! ~~cpascual 20121121
            print 'TaurusDevTree.%s: %s' % (self.getLogLevel(), msg)

    def setTangoHost(self, tango_host=None):
        self.db = taurus.Authority(tango_host)

    # model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel,
        # TaurusBaseWidget.setModel,
        # TaurusBaseWidget.resetModel)

    def getModel(self):
        return self._filters

    def getModelClass(self):
        return list  # taurus.core.taurusauthority.TaurusAuthority

    def setModel(self, model):
        TaurusBaseWidget.setModel(self, model)

    def setModelCheck(self, model):
        # Called from TaurusBaseWidget.setModel()
        self.trace('setModelCheck(%s)' % str(model)[:80])
        self.loadTree(model)

    @Qt.pyqtSlot('QStringList')
    def addModels(self, modelNames):
        '''Adds models to the existing ones:
        :param modelNames:  (sequence<str>) the names of the models to be added
        .. seealso:: :meth:`removeModels`
        '''
        self.trace('In addModels(%s)' % str(modelNames)[:80])
        modelNames = split_model_list(modelNames)
        self.setTree(self.getTangoDict(modelNames), clear=False)
        if isSequence(modelNames):
            self._filters = sorted(
                set(split_model_list(self._filters) + modelNames))
        elif isMap(modelNames):
            if isMap(self._filters):
                self._filters.update(modelNames)
            else:
                self._filters = modelNames

    ##########################################################################
    # Loading/Cleaning the tree

    # def loadTree(self,filters,clear=False):
        #'''
        # This method show a list of instances and devices depending on the given servers in QTProperty or in another widget,
        # this method can be used to connect TauDevTree with another widget such as LineEdit.
        #'''
        #self.trace('In loadTree(%s)'%str(filters))
        #if clear: self.setWindowTitle('TaurusDevTree:%s'%str(filters))
        # self.setTree(self.getTangoDict(filters),clear=clear,alias=False)

    def loadTree(self, filters):
        try:
            if isString(filters):
                try:
                    assert '{' in filters
                    filters = dict(filters)
                except:
                    filters = split_model_list(filters)
            self.trace('loadTree(%s)' % (filters))
            assert isMap(filters) or isSequence(
                filters), "Filters have to be map, string or list type!"
            # self._filters = filters
            properties.set_property(self, 'Filters', filters)
            if isSequence(filters):
                self.setWindowTitle('TaurusDevTree:%s' % str(filters))
                dct = self.getTangoDict(filters)
            else:  # if isMap(filters):
                self.setWindowTitle('TaurusDevTree:%s' %
                                    ','.join(filters.keys()))

                def expand_dict(d):
                    return [x for v in d.values() for x in (expand_dict(v) if hasattr(v, 'values') else (v,))]
                targets = [t.upper() for t in get_matching_devices(
                    ['*%s*' % f if '*' not in f else f for f in expand_dict(filters)])]

                def get_devs(f):
                    return dict.fromkeys(t for t in targets if matchCl(f, t))

                def expand_filter(f):
                    return dict((k, expand_filter(v) if hasattr(v, 'values') else get_devs(v)) for k, v in f.items() if v)
                dct = expand_filter(filters)
            # self.Loader.next([self.setTree,dct,True])
            self.setTree(dct, clear=True)
        except:
            self.warning('TaurusDeviceTree.loadTree(%s):\n%s' %
                         (filters, traceback.format_exc()))

    def setTree(self, diction, clear=False):
        """
        Initializes the tree from a dictionary {'Node0.0':{'Node1.0':None,'Node1.1':None}}
        """
        K = len(str(dict(diction)))
        self.trace('In setTree(%d) ...' % K)
        self.debug(diction)
        if diction is None:
            return
        if clear or self.dictionary:
            if not clear:
                # Merging new and old models
                diction = djoin(diction, self.dictionary)
            self.clear()
        self.dictionary = diction
        if len(diction):
            self.setNodeTree(self, diction, alias=(
                self.getShowAlias() or K < self.getMaxDevices() * 20))
            # Auto-Expand caused problems when loading filters from QSettings
            if 0 < len(self.item_list) < self.getMaxDevices():
                self.expandAll(queue=False)

    def setNodeTree(self, parent, diction, alias=False):
        """
        It has parent as argument to allow itself to be recursive
        Initializes the node tree from a dictionary {'Node0.0':{'Node1.0':None,'Node1.1':None}}
        """
        self.debug('In setNodeTree(%d,alias=%s) ...' % (len(diction), alias))
        if not hasattr(diction, 'keys'):
            diction = dict.fromkeys(diction)
        for node in sorted(diction.keys()):
            assert int(self.index) < 10000000000, 'TooManyIterations!'
            self.index = self.index + 1
            dev_alias = alias and str(node).count(
                '/') == 2 and get_alias_for_device(node)
            text = '%s (%s)' % (node, dev_alias) if dev_alias else node
            if diction[node] and any(diction[node]):
                item = self.createItem(parent, node, text)
                self.setNodeTree(item, diction[node], alias)
            else:
                item = self.createItem(parent, node, text)

    def clear(self):
        while not self.Expander.getQueue().empty():
            self.Expander.getQueue().get()
        self.item_index.clear()
        while self.item_list:
            self.item_list.pop()
        Qt.QTreeWidget.clear(self)

    def refreshTree(self):
        self.loadTree(self._filters)
        self.refreshTree.emit()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # @name Methods for building server/devices/attributes tree
    # @{
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getTangoDict(self, filters):
        self.trace('In TaurusDevTree.getTangoDict(%s(%s))' %
                   (type(filters), str(filters)[:80]))
        if filters is None:
            return
        result = {}
        filters = split_model_list(filters)
        targets = get_matching_devices(filters)
        targets = [t.upper() for t in targets]
        domains = set(t.split('/')[0] for t in targets)
        for d in domains:
            families = set(t.split('/')[1]
                           for t in targets if t.startswith('%s/' % d))
            result[d] = dict((f, dict.fromkeys(
                t for t in targets if t.startswith('%s/%s/' % (d, f)))) for f in families)
        return result

    def addAttrToDev(self, my_device, expert=False, allow_types=None):
        """ This command returns the list of attributes of a given device applying display level and type filters.
        @argin expert If False only PyTango.DispLevel.OPERATOR attributes are displayed
        @argin allow_types Only those types included in the list will be displayed (e.g. may be restricted to numeric types only)
        """
        numeric_types = [PyTango.DevDouble, PyTango.DevFloat, PyTango.DevLong, PyTango.DevLong64,
                         PyTango.DevULong, PyTango.DevShort, PyTango.DevUShort, PyTango.DevBoolean, PyTango.DevState]
        allow_types = allow_types or [PyTango.DevString] + numeric_types
        dct = {}
        self.trace('In addAttrToDev(%s)' % my_device)
        try:
            proxy = PyTango.DeviceProxy(my_device)
            timeout = proxy.get_timeout_millis()
            proxy.set_timeout_millis(50)
            proxy.ping()
            list_attr = proxy.attribute_list_query()
            proxy.set_timeout_millis(timeout)

            for aname, my_attr in sorted([(a.name, a) for a in list_attr]):
                if allow_types and my_attr.data_type not in allow_types:
                    continue
                if not expert and my_attr.disp_level == PyTango.DispLevel.EXPERT:
                    continue
                label = aname == my_attr.label and aname.lower(
                ) or "%s (%s)" % (aname.lower(), my_attr.label)
                dct[str(my_device).lower() + '/' + label] = 0
        except PyTango.DevFailed, e:
            self.warning('addAttrToDev(%s): %s' % (my_device, str(e)))
            qmsg = Qt.QMessageBox(Qt.QMessageBox.Critical, '%s Error' %
                                  my_device, '%s not available' % my_device, Qt.QMessageBox.Ok, self)
            qmsg.show()
        except Exception, e:
            self.warning('addAttrToDev(%s): %s' % (my_device, str(e)))
            qmsg = Qt.QMessageBox(Qt.QMessageBox.Critical, '%s Error' %
                                  my_device, str(e), Qt.QMessageBox.Ok, self)
            qmsg.show()
        return dct

    def addAttrToNode(self, node=None, full=False):
        node = node or self.currentItem()
        dev = self.getNodeDeviceName(node)
        self.trace('In addAttrToNode(%s)' % dev)
        attrs = self.addAttrToDev(dev)
        children = [str(node.child(i).text(0)).lower()
                    for i in range(node.childCount())]
        for aname in sorted(attrs):
            tag = aname.rsplit('/')[-1]
            if tag.lower() in children:
                continue
            elif not full and not self.matchAttrFilter(aname):
                continue
            else:
                natt = self.createItem(node, value=aname, text=tag)
                natt.draggable = aname.split()[0].strip()
                natt.isAttribute = True
                natt.DeviceName = dev
                icon = self.getNodeIcon(natt)
                if icon:
                    natt.setIcon(0, icon)
                # it gets all aliases for this device attributes
                alias = getattr(node, 'AttributeAlias', {})
                if alias:
                    self.trace('Got aliases for %s: %s' % (aname, alias))
                    [setattr(natt, 'AttributeAlias', v)
                     for k, v in alias.items() if k in aname.lower()]
                else:
                    natt.AttributeAlias = aname.split()[0].strip()
        node.setExpanded(True)
        return

    ###########################################################################
    # Node getters

    def getNode(self, target=None):
        """ Gets currrent node or node by name or by regexp """
        if target is None:
            return self.currentItem()
        else:
            nodes = self.getMatchingNodes(target, 1)
            if not nodes:
                return None
            else:
                return nodes[0]
        return

    def getNodeByName(self, key):
        return self.item_index[key]

    def getNodeList(self):
        return self.item_index.keys()

    def getMatchingNodes(self, regexp, limit=0, all=False, exclude=None):
        """ It returns all nodes matching the given expression. """
        result, regexp = [], str(regexp).lower()
        exclude = exclude or []
        self.trace('In TauDevTree.getMatchingNodes(%s,%s,%s,%s)' %
                   (regexp, limit, all, exclude))
        if not all:
            node = self.item_index.get(regexp, None)
            if node is not None:
                return [node]
        regexp = re.compile(extend_regexp(regexp))
        for k, node in self.item_index.iteritems():
            nname = self.getNodeText(node, full=True).lower()
            if (regexp.match(k) or regexp.match(nname)) and \
                    (not exclude or not any(re.match(x.lower(), y) for x in exclude for y in (k.lower(), nname))):
                result.append(node)
                if not all and len(result) == 1:
                    break
                if limit and len(result) >= limit:
                    break
        return result

    def getSelectedNodes(self):
        return self.selectedItems()

    def getAllNodes(self):
        """ Returns a list with all node objects. """
        def get_child_nodes(dct, node, fun=None):
            if fun:
                fun(node)
            dct.update([(str(node.text(0)), node)])
            for j in range(node.childCount()):
                get_child_nodes(dct, node.child(j))
            return dct
        dct = {}
        for i in range(self.topLevelItemCount()):
            get_child_nodes(dct, self.topLevelItem(i))
        return dct

    def unpackChildren(self):
        """ removes all nodes from the tree and returns them in a list, used for resorting """
        allChildren = []
        nodes = self.getAllNodes().values()

        for node in nodes:
            allChildren.extend(node.takeChildren())
        while self.topLevelItemCount():
            allChildren.append(self.takeTopLevelItem(0))
        return allChildren

    # @}
    ##########################################################################

    ###########################################################################
    # Expand/Collapse/Search nodes

    def collapseNode(self, ALL=False, filters='', fun=None):
        """ Collapses the whole tree or from a given node.
        @argin ALL tells whether to collapse from current item or the whole tree
        @argin filters Allows to set a list of nodes to not be filtered

        """
        filters = str(filters).lower()
        found = ''
        self.debug('In TaurusTree.collapseAll(%s)' % filters)
        todelete = []

        def expand_child_nodes(node):
            result = int(bool(filters))
            if fun:
                fun(node)
            if not node:
                return ''
            for j in range(node.childCount()):
                child = node.child(j)
                result = expand_child_nodes(child)
                if filters and re.search(filters, str(child.text(0)).lower()):
                    self.debug('In TaurusTree.collapseAll(%s): %s matches!' % (
                        filters, str(child.text(0)).lower()))
                    result = True
                elif not result:
                    child.setExpanded(False)
                # When collapsing all attribute lists are cleaned up
                aname = '/'.join(['[0-9a-zA-Z\-\_]+'] * 4)
                if re.match(aname, str(child.text(0))):
                    todelete.append((node, child))
            if not result:
                node.setExpanded(False)
            return result
        if ALL:
            for i in range(self.topLevelItemCount()):
                found = expand_child_nodes(self.topLevelItem(i)) or found
        else:
            found = expand_child_nodes(self.currentItem()) or found
        for node, child in todelete:  # Pruning attribute nodes
            node.removeChild(child)
            del child
        return found

    ###########################################################################
    # New expand/search methods

    def expandNode(self, node=None, expand=True):
        """ Needed to do threaded expansion of the tree """
        if node is None:
            node = self.getNode()
        if isinstance(node, (basestring, Qt.QString)):
            name, node = str(node), self.getNode(node)
        else:
            name = self.getNodeText(node)
        node.setExpanded(expand)
        return expand

    def expandAll(self, queue=True):
        self.findInTree('*', select=False, queue=queue)

    def findDialog(self):
        self.findInTree(str(Qt.QInputDialog.getText(
            self, 'Search ...', 'Write a part of the name', Qt.QLineEdit.Normal)[0]))

    @Qt.pyqtSlot('QString')
    def findInTree(self, regexp, collapseAll=None, exclude=None, select=True, queue=True):
        self.trace('In TauTree.findInTree(%s)' % regexp)
        if collapseAll is None:
            collapseAll = self.collapsing_search
        regexp = str(regexp).lower().strip()
        exclude = (lambda x: x if hasattr(x, '__iter__') else [x])(
            exclude or self.excludeFromSearch or [])
        if not regexp:
            return
        try:
            t0 = time.time()
            nodes = self.getMatchingNodes(regexp, all=True, exclude=exclude)
            if len(nodes) > 150:
                v = Qt.QMessageBox.warning(None, 'Device Tree Search',
                                           'Your search matches too many devices (%d) and may slow down the application.\nDo you want to continue?' % len(
                                               nodes),
                                           Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel)
                if v == Qt.QMessageBox.Cancel:
                    self.debug('Search cancelled by user.')
                    return
            if nodes:
                # It's good to have first node matched to be selected fast
                if select:
                    nodes[0].setSelected(True)
                    self.setCurrentItem(nodes[0])
                    # Searches must not trigger events!
                    self.deviceSelected(self.getNodeDeviceName(nodes[0]))
                    self.debug('The selected node is %s' %
                               self.getNodeText(nodes[0]))
                # Then proceed to expand/close the rest of nodes
                parents = set(
                    parent for node in nodes for parent in self.getNodePath(node) if parent)
                for item in self.item_list:
                    matched, expanded = item in parents, item.isExpanded()
                    if (matched and not expanded):
                        if queue:
                            self.Expander.getQueue().put((item, True))
                        else:
                            item.setExpanded(True)
                    elif (not matched and expanded and self.collapsing_search):
                        if queue:
                            self.Expander.getQueue().put((item, False))
                        else:
                            item.setExpanded(False)
                if select:
                    self.scrollTo(self.indexFromItem(
                        nodes[0]), Qt.QAbstractItemView.PositionAtTop)  # Center)
                self.debug('\tfindInTree(%s): %d nodes found in %f s' %
                           (regexp, len(nodes), time.time() - t0))
            else:
                if collapseAll:
                    if queue:
                        [self.Expander.getQueue().put((item, False))
                         for item in self.item_list if item.isExpanded()]
                    else:
                        [item.setExpanded(False)
                         for item in self.item_list if item.isExpanded()]
                self.debug('findInTree(%s): Node not found' % (regexp))
            if queue:
                self.Expander.next()
        except:
            self.warning('findInTree(%s): failed' % (regexp))
            self.error(traceback.format_exc())

    def sortCustom(self, order):
        assert order and len(order), 'sortCustom(order) must not be empty'
        allChildren = {}
        while self.topLevelItemCount():
            it = self.takeTopLevelItem(0)
            allChildren[str(it.text(0))] = it

        sorter = lambda k, ks=[re.compile(c) for c in order]: str(
            (i for i, r in enumerate(ks) if r.match(k.lower())).next()) + str(k)
        for c, it in sorted(allChildren.items(), key=lambda k: sorter(k[0])):
            self.debug('tree.sortCustom(%s): %s inserted at %d' %
                       (order, it.text(0), self.topLevelItemCount()))
            self.insertTopLevelItem(self.topLevelItemCount(), it)
        return

    ###########################################################################
    # Update node colors

    def setIcons(self, dct={}, root_name=None, regexps=True):
        '''
        This method change the icons depending of the status of the devices
        Dict is a dictionary with name of device and colors such as {name_device:color,name_device2:color2}
        An alternative may be an icon name!
        '''
        #secs = time.time()
        #ID = int(100*random.random())
        state2color = lambda state: Qt.QColor(
            DEVICE_STATE_PALETTE.number(state))
        #quality2color = lambda attr: Qt.QColor(ATTRIBUTE_QUALITY_PALETTE.number(quality))

        def update_node(node, key, dct):
            if hasattr(node, 'CustomForeground'):
                node.setForeground(0, Qt.QBrush(
                    Qt.QColor(node.CustomForeground)))
            if hasattr(node, 'CustomBackground'):
                node.setBackground(0, Qt.QBrush(
                    Qt.QColor(node.CustomBackground)))
            elif hasattr(node, 'StateBackground'):
                node.setBackground(0, Qt.QBrush(state2color(dct[key])))
            if hasattr(node, 'CustomIcon'):
                node.setIcon(0, Qt.QIcon(node.CustomIcon))
            else:
                #key = str(node.text(0)).split(' ')[0]
                if key.count('/') == 2:
                    self.setStateIcon(node, dct and dct[key] or '')
            return

        if not isinstance(dct, dict):
            dct = dict.fromkeys(dct, '')
        nodes = self.getAllNodes()
        for name, node in nodes.iteritems():
            name = str(name).split()[0]
            if node.isHidden():
                continue
            if regexps:
                matches = [v for k, v in dct.items() if re.match(
                    k.lower(), name.lower())]
                if matches:
                    update_node(node, name, {name: matches[0]})
            elif name in dct or not dct:
                update_node(node, name, dct or {name: ''})
        return

    def setStateIcon(self, child, color):
        if icons_dev_tree is None:
            self.debug('In setStateIcon(...): Icons for states not available!')
            return
        if color == "#00ff00" or color in 'ON,OPEN,EXTRACT':
            icon = Qt.QIcon(":/ICON_GREEN")
            child.setIcon(0, icon)
        elif color == "#ff0000" or color in 'OFF,FAULT':
            icon = Qt.QIcon(":/ICON_RED")
            child.setIcon(0, icon)
        elif color == "#ff8c00" or color in 'ALARM':
            icon = Qt.QIcon(":/ICON_ORANGE")
            child.setIcon(0, icon)
        elif color == "#ffffff" or color in 'CLOSE,INSERT':
            icon = Qt.QIcon(":/ICON_WHITE")
            child.setIcon(0, icon)
        elif color == "#80a0ff" or color in 'MOVING,RUNNING':
            icon = Qt.QIcon(":/ICON_BLUE")
            child.setIcon(0, icon)
        elif color == "#ffff00" or color in 'STANDBY':
            icon = Qt.QIcon(":/ICON_YELLOW")
            child.setIcon(0, icon)
        elif color == "#cccc7a" or color in 'INIT':
            icon = Qt.QIcon(":/ICON_BRAWN")
            child.setIcon(0, icon)
        elif color == "#ff00ff" or color in 'DISABLE':
            icon = Qt.QIcon(":/ICON_PINK")
            child.setIcon(0, icon)
        elif color == "#808080f" or color in 'UNKNOWN':
            icon = Qt.QIcon(":/ICON_GREY")
            child.setIcon(0, icon)
        else:
            icon = Qt.QIcon(":/ICON_WHITE")
            child.setIcon(0, icon)

    ##########################################################################
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Event methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def deviceClicked(self, item, column):
        self.trace("In TaurusDevTree.deviceClicked(%s)" % item.text(column))
        self.deviceSelected(self.getNodeDeviceName())

    def deviceSelected(self, device_name=''):
        '''QSIGNAL: this method is used to emit deviceSelected(QString) signal'''
        self.trace("In TaurusDevTree.deviceSelected(%s)" % device_name)
        try:
            #item = self.currentItem()
            device_name = device_name or self.getNodeDeviceName()  # item.text(0)
            if str(device_name).count('/') != 2:
                return
            # Signal
            self.trace('TaurusTree emit deviceSelected(%s) signal ...' %
                       device_name)
            self.deviceSelected.emit(Qt.QString(device_name))
        except:
            self.error(traceback.format_exc())
            pass

    def getModelMimeData(self):
        '''Returns a MimeData object containing the model data. The default implementation
        fills the `TAURUS_MODEL_MIME_TYPE`. If the widget's Model class is
        Attribute or Device, it also fills `TAURUS_ATTR_MIME_TYPE` or
        `TAURUS_DEV_MIME_TYPE`, respectively
        '''
        mimeData = Qt.QMimeData()
        node = self.currentItem()
        draggable = self.getNodeDraggable(node)
        if draggable:
            slashes = draggable.count('/') - draggable.count(':')
            # mimeData.setData('application/x-qabstractitemmodeldatalist',draggable)
            if slashes == 3:
                mimeData.setData(TAURUS_ATTR_MIME_TYPE, draggable)
            elif slashes == 2:
                mimeData.setData(TAURUS_DEV_MIME_TYPE, draggable)
            else:
                mimeData.setData(TAURUS_MODEL_MIME_TYPE, draggable)
        return mimeData

    def checkHeaderClicked(self, position):
        if self.itemAt(position) is self.headerItem():
            node = self.headerItem()
            self.showNodeContextMenu(node, event)
            #node.ContextMenu = ['Search ...']

    def mouseMoveEvent(self, event):
        '''
        copied from TaurusBaseWidget to provide drag events
        It had to be rewritten as QTreeWidget does not allow drag events
        '''
        self.debug('In TaurusDevTree.mouseMoveEvent')
        if not self._dragEnabled or not event.buttons() & Qt.Qt.LeftButton:
            return self.getQtClass().mouseMoveEvent(self, event)
        if (event.pos() - self.dragStartPosition).manhattanLength() < Qt.QApplication.startDragDistance():
            return self.getQtClass().mouseMoveEvent(self, event)
        # The mouseMoveEvent of QTreeWidget do not allow drag, commented
        ret = None  # self.getQtClass().mouseMoveEvent(self, event) #call the superclass
        event.accept()  # we make sure we accept after having called the superclass so that it is not propagated (many default implementations of mouseMoveEvent call event.ignore())
        drag = Qt.QDrag(self)
        drag.setMimeData(self.getModelMimeData())
        drag.exec_(Qt.Qt.CopyAction, Qt.Qt.CopyAction)
        return ret

    def mimeTypes(self):
        return self.getSupportedMimeTypes()

    def dropEvent(self, event):
        '''reimplemented to support dropping of modelnames in forms'''
        self.debug('dropEvent(%s): %s,%s' % (event, event.mimeData(),
                                             split_model_list(event.mimeData().formats())))
        if event.source() is self:
            self.info('Internal drag/drop not allowed')
            return
        if any(s in event.mimeData().formats() for s in self.getSupportedMimeTypes()):
            # mtype =
            # self.handleMimeData(event.mimeData(),self.addModels)#lambda
            # m:self.addModels('^%s$'%m))
            event.acceptProposedAction()
        else:
            self.warning('Invalid model in dropped data')

    # 3
    # @name Context Menu Actions
    # @{

    def contextMenuEvent(self, event):
        '''
        This function is called when right clicking on TaurusDevTree area.
        '''
        node = self.currentItem()
        self.showNodeContextMenu(node, event)
        return

    def showNodeContextMenu(self, node, event):
        """
        A pop up menu will be shown with the available options.
        Menus are managed using two tuple lists for each node: node.ContextMenu and node.ExpertMenu
        """
        obj = self.getNodeDraggable(node)
        position = event.globalPos()
        self.debug('showNodeContextMenu(%s)' % obj)
        if self.itemAt(position) is self.headerItem():
            node = self.headerItem()
            #node.ContextMenu = ['Search ...']
        if node is None:
            node = self
        else:
            if not hasattr(node, 'ContextMenu'):
                node.ContextMenu = []
            # Creating default menu
            if not 'Search ...' in [k for k, a in node.ContextMenu]:
                # DEVICE NODE CONTEXT MENU
                if obj.count('/') == 2:

                    node.ContextMenu.append(("Open Panel", self.showPanel))
                    node.ContextMenu.append(
                        ("Show Attributes", self.addAttrToNode))
                    if self.getNodeAdmin(node):
                        node.ContextMenu.append(("Go to %s" % self.getNodeAdmin(node),
                                                 lambda p=self.getNodeAdmin(
                                                     node): p and self.findInTree(p)
                                                 ))
                    if not hasattr(node, 'ExpertMenu'):
                        setattr(node, 'ExpertMenu', self.ExpertMenu)  # [])
                    if not 'Show Properties' in [k for k, a in node.ExpertMenu]:
                        node.ExpertMenu.append(
                            ("Show Properties", self.showProperties))

                        def test_device():
                            device = str(self.getNodeDeviceName())
                            if device:
                                comm = 'tg_devtest %s &' % device
                                os.system(comm)
                            else:
                                self.debug(
                                    'TaurusDevTree.TestDevice: Selected Device is None!')
                        node.ExpertMenu.append(("Test Device", test_device))
                        node.ExpertMenu.append(
                            ("Show ALL Attributes", lambda s=self: s.addAttrToNode(full=True)))
                    node.ContextMenu.append(('', None))

                # ATTRIBUTE NODE CONTEXT MENU
                elif obj.count('/') == 3:
                    for k, v in self.AttributeMenu:
                        self.debug('Adding action %s' % k)
                        if type(v) is str and hasattr(self, v):
                            node.ContextMenu.append((k, getattr(self, v)))
                        else:
                            node.ContextMenu.append(
                                (k, lambda s=self.getNodeAlias(node): v(s)))
                    #node.ContextMenu.append(("add to Trends", self.addToPlot))
                    #node.ContextMenu.append(("remove from Trends", self.removeFromPlot))
                    node.ContextMenu.append(('', None))
                elif not hasattr(node, 'ExpertMenu'):
                    setattr(node, 'ExpertMenu', self.ExpertMenu)  # [])
                #node.ContextMenu.append(("Expand Node", self.expandNode))
                #node.ContextMenu.append(("Collapse Node", self.collapseNode))
                if node.isExpanded() and node.childCount() < 10 and all(self.getNodeText(node.child(j)).count('/') == 2 for j in range(node.childCount())):
                    node.ContextMenu.append(("Show Attributes", lambda n=node, s=self: [
                                            s.addAttrToNode(n.child(j)) for j in range(n.childCount())]))
                node.ContextMenu.append(("Search ...",
                                         lambda: self.findInTree(str(Qt.QInputDialog.getText(
                                             self, 'Search ...', 'Write a part of the name', Qt.QLineEdit.Normal)[0]))
                                         ))
        #configDialogAction = menu.addAction("Refresh Tree")
        #self.connect(configDialogAction, Qt.SIGNAL("triggered()"), self.refreshTree)
        menu = Qt.QMenu(self)

        if hasattr(node, 'ContextMenu'):
            last_was_separator = True
            for t in (type(node.ContextMenu) is dict and node.ContextMenu.items() or node.ContextMenu):
                try:
                    k, action = t
                    if k:
                        configDialogAction = menu.addAction(k)
                        if action:
                            configDialogAction.triggered.connect(action)
                        else:
                            configDialogAction.setEnabled(False)
                        last_was_separator = False
                    elif not last_was_separator:
                        menu.addSeparator()
                        last_was_separator = True
                except Exception, e:
                    self.warning('Unable to add Menu Action: %s:%s' % (t, e))

        if hasattr(node, 'ExpertMenu'):
            menu.addSeparator()
            expert = menu.addMenu('Expert')
            # expert.addSeparator()
            last_was_separator = True
            for t in (type(node.ContextMenu) is dict and node.ExpertMenu.items() or node.ExpertMenu):
                try:
                    k, action = t
                    if k:
                        configDialogAction = expert.addAction(k)
                        if action:
                            configDialogAction.triggered.connect(action)
                        else:
                            configDialogAction.setEnabled(False)
                        last_was_separator = False
                    elif not last_was_separator:
                        expert.addSeparator()
                        last_was_separator = True
                except Exception, e:
                    self.warning('Unable to add Expert Action: %s:%s' % (t, e))
        # menu.addSeparator()
        menu.exec_(event.globalPos())
        del menu


class PopupDialog(Qt.QDialog):
    """
    This class create the dialog
    Dialog is used to make new floating panels persistent
    """

    def __init__(self, parent=None, target=None):
        Qt.QDialog.__init__(self, parent)
        if target:
            self.initComponents(target)

    def initComponents(self, newWidget, show=True):
        widgetLayout = Qt.QVBoxLayout(self)
        widgetLayout.setContentsMargins(10, 10, 10, 10)
        widgetLayout.addWidget(newWidget)
        self.setWindowTitle(newWidget.windowTitle())
        self.setModal(False)
        self.setVisible(True)
        if show:
            self.exec_()

##########################################################################


class TaurusTreeNode(Qt.QTreeWidgetItem, TaurusBaseComponent):
    """
    Unused; one day should replace TaurusTreeNodeContainer and all methods moved here.
    Base class for all Taurus Tree Node Items
    """

    #-------------------------------------------------------------------------
    # Write your own code here to define the signals generated by this widget
    #

    modelChanged = Qt.pyqtSignal('const QString &')

    def __init__(self, name=None, parent=None):
        name = name or self.__class__.__name__
        self.call__init__wo_kw(Qt.QTreeWidgetItem, parent)
        self.call__init__(TaurusBaseComponent, name, parent)
        # self.defineStyle()

    # def defineStyle(self):
        #""" Defines the initial style for the widget """
        # -----------------------------------------------------------------------
        # Write your own code here to set the initial style of your widget
        ##
        # self.updateStyle()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory methods to be implemented in any subclass of TaurusComponent
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getParentTaurusComponent(self):
        """ Returns a parent Taurus component or None if no parent TaurusBaseComponent
            is found."""
        p = self.parentItem()
        while p and not isinstance(p, TaurusTreeNode):
            p = self.parentItem()
        return p

    def updateStyle(self):
        """ Method called when the component detects an event that triggers a change
            in the style."""

        # if self.scene():
        #    self.scene().updateSceneItem(self)

        #----------------------------------------------------------------------
        # Write your own code here to update your widget style

        # send a repaint in the end
        # self.repaint()
        #if self._parent: self._parent.repaint()

        state2color = lambda state: Qt.QColor(
            DEVICE_STATE_PALETTE.number(state))
        quality2color = lambda attr: Qt.QColor(
            ATTRIBUTE_QUALITY_PALETTE.number(attr))
        v = self.getModelValueObj()
        if isinstance(v, PyTango.DevState):  # TODO: maybe change to Taurus.tango.DevState
            # @TODO: node is undefined. Check code
            node.setBackground(0, Qt.QBrush(state2color(v)))
        if hasattr(v, 'quality'):
            self.setForeground(0, Qt.QBrush(quality2color(v.quality)))

    def isReadOnly(self):
        return True

    def __str__(self):
        return self.log_name + "(" + self.modelName + ")"

    def getModelClass(self):
        return taurus.core.taurusdevice.TaurusDevice

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseComponent over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    # def attach(self):
        #"""Attaches the widget to the model"""
        # if self.isAttached():
        # return True
        # -----------------------------------------------------------------------
        # Write your own code here before attaching widget to attribute connect
        # the proper signal so that the first event is correctly received by the
        # widget
        ##
        # Typical code is:
        # self.connect(self, Qt.SIGNAL('valueChangedDueToEvent(QString)'),
        # self.setTextValue)
        #ret = TaurusBaseWidget.attach(self)
        # by default enable/disable widget according to attach state
        # self.setEnabled(ret)
        # return ret

    # def detach(self):
        #"""Detaches the widget from the model"""
        # TaurusBaseWidget.detach(self)

        # -----------------------------------------------------------------------
        # Write your own code here after detaching the widget from the model
        ##
        # Typical code is:
        # self.emit(Qt.SIGNAL('valueChangedDueToEvent(QString)'),
        # Qt.QString(value_str))
        # self.disconnect(self, Qt.SIGNAL('valueChangedDueToEvent(QString)'),
        # self.setTextValue)
        # by default disable widget when dettached
        # self.setEnabled(False)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel,
                            TaurusBaseWidget.setModel,
                            TaurusBaseWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                     TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel,
                                     TaurusBaseWidget.resetUseParentModel)

    #-------------------------------------------------------------------------
    # Write your own code here for your own widget properties


class TaurusDevTreeOptions(Qt.QWidget):
    """ This class provides a search(QString) signal to be connected to TaurusDevTree.findInTree slot """

    search = Qt.pyqtSignal('QString')
    loadTree = Qt.pyqtSignal('QString')
    hideUnarchived = Qt.pyqtSignal()
    hideUnexported = Qt.pyqtSignal()

    def __init__(self, parent=None, icon=None):
        Qt.QWidget.__init__(self, parent)

        self.setLayout(Qt.QHBoxLayout())
        try:
            self._pixmap = Qt.QPixmap(icon or 'image/icons/search.png')
            self._label = Qt.QLabel(self)
            self._label.setPixmap(self._pixmap)
            self.layout().addWidget(self._label)
        except:
            pass

        self._edit = Qt.QLineEdit()
        self._button = Qt.QPushButton()
        self._button.setText('Search')
        self._edit.returnPressed.connect(self._button.animateClick)
        self._button.clicked.connect(self._emitSearch)
        self.layout().addWidget(self._edit)
        self.layout().addWidget(self._button)

    def connectWithTree(self, tree):
        self.search.connect(tree.findInTree)

    def _emitSearch(self):
        text = self._edit.text()
        if text:
            self.search.emit(text)
        return

SearchEdit = TaurusDevTreeOptions

##########################################################################


class ServerBrowser(TaurusDevTree):
    """ This class is used only when browsing by Server/Instance instead of Domain/Family/Member scheme """

    def getDeviceDict(self, filters):
        '''
        This method build a dictionary of instances and devices depending on the given servers,devices or instances in QTProperty or in another widget
        --- filters is a string with names of devices/servers such as "LT/VC/ALL,LT02/VC/IP-01" or "modbus,pyplc"
        --- filters is a list of devices such as ["LT/VC/ALL","LT02/VC/IP-01"]
        '''
        self.trace('In TaurusDevTree.buildDictFromFilters(%s)' % filters)
        self._filters = filters
        # @TODO:QString and QStringList should not be used (They disappear in API2)
        if type(filters) == type("") or isinstance(filters, Qt.QString):
            filters = str(filters).split(',')
        # @TODO:QString and QStringList should not be used (They disappear in API2)
        elif isinstance(filters, Qt.QStringList):
            filters = list(filters)
        elif type(filters) != type([]):
            self.debug("'filters' has to be a string or the list type!")
        vals = {}
        if not filters:
            return vals
        if filters[0].count('/') == 0:
            self.debug(
                'In TaurusDevTree.buildDictFromFilters(%s): Getting Servers' % filters)
            targets, addMe = self.db.get_server_name_list(
            ), self.addInstToServ  # Searching servers
        elif filters[0].count('/') == 1:
            self.debug(
                'In TaurusDevTree.buildDictFromFilters(%s): Getting Instances' % filters)
            targets, addMe = self.db.get_server_list(), self.addDevToInst  # Searching instances
        elif filters[0].count('/') == 2:
            self.debug(
                'In TaurusDevTree.buildDictFromFilters(%s): Getting Devices' % filters)
            targets, addMe = self.db.get_device_exported(
                "*"), lambda s: {s: {}}  # self.addAttrToDev #Searching devices
        else:
            raise Exception('UnknownFilter!: %s' % filters[0])

        for t in targets:
            for f in filters:
                f = str(f)
                exp = f.replace(
                    '*', '.*').lower() if '*' in f and '.*' not in f else f.lower()
                if re.match(exp, t.lower()):
                    self.debug('Adding node %s' % t)
                    vals[t] = addMe(t)
        self.trace('Out of TaurusDevTree.getDeviceDict(%s)' % filters)
        return vals

    def addInstToServ(self, my_server):
        dict = {}
        list_inst = self.get_instances_for_server(my_server)
        lower_list_inst = [s.lower() for s in list_inst]
        for my_inst in lower_list_inst:
            if self._expand:
                dict[my_inst] = self.addDevtoInst(my_inst)
            else:
                dict[my_inst] = 0
        return dict

    def addDevtoInst(self, my_inst, expand_attrs=False):
        d = {}
        list_dev = self.get_devices_for_instance(my_inst)
        lower_list_dev = [s.lower() for s in list_dev]
        for my_dev in lower_list_dev:
            if self._expand:
                d[my_dev] = self.addAttrToDev(
                    my_dev) if expand_attrs else {my_dev: {}}
            else:
                d[my_dev] = 0
        return d

    def addFamilyToDomain(self, prev, expand_attrs):
        d = {}
        #children = self.get_devices_for_instance(my_inst)
        # @TODO: list_dev is undefined. Check code
        lower_list_dev = [s.lower() for s in list_dev]
        for my_dev in lower_list_dev:
            if self._expand:
                d[my_dev] = self.addAttrToDev(
                    my_dev) if expand_attrs else {my_dev: {}}
            else:
                d[my_dev] = 0
        return d

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Methods  for database commands
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def get_instances_for_server(self, server_name):
        #executable_name = class_name
        instances = self.db.get_instance_name_list(server_name)
        return [server_name + '/' + instance for instance in instances]

    def get_devices_for_instance(self, instance_name):
        devslist = self.db.get_device_class_list(instance_name)
        return [dev for dev in devslist if '/' in dev and not dev.startswith('dserver')]

    # @}

##########################################################################


class TaurusSearchTree(TaurusWidget):
    """
    This Class is a mere wrapper providing a TaurusDevTree connected with an options panel and some optional checkboxes.
    """
    __slots__ = (
        "setTangoHost",
        #"setModel",
        "addModels",
        "setModelCheck",
        "setTree",
        "findInTree",
        "expandAll",
        "loadTree",
    )

    refreshTree = Qt.pyqtSignal()
    nodeFound = Qt.pyqtSignal()
    deviceSelected = Qt.pyqtSignal('QString')
    addAttrSelected = Qt.pyqtSignal('QStringList')
    removeAttrSelected = Qt.pyqtSignal('QStringList')
    modelChanged = Qt.pyqtSignal('const QString &')
    search = Qt.pyqtSignal('QString')

    @staticmethod
    def method_forwarder(*args, **kwargs):
        m, o = kwargs['method'], kwargs['object']
        # print 'Calling %s.%s'%(o,m)
        getattr(o.__class__, m)(o, *args)

    def defineStyle(self):
        #print('In TaurusSearchTree.defineStyle()')
        self.setWindowTitle('TaurusDevTree')
        self.setLayout(Qt.QVBoxLayout())
        self.edit = TaurusDevTreeOptions(self)
        self.tree = TaurusDevTree(self)
        self.layout().addWidget(self.edit)
        self.layout().addWidget(self.tree)
        self.registerConfigDelegate(self.tree)
        # Slot forwarding ...
        for k in TaurusDevTree.__dict__.keys():
            # if k in ['__init__','defineStyle']: continue
            if k not in self.__slots__:
                continue
            try:
                setattr(self, k, partial(
                    self.method_forwarder, method=k, object=self.tree))
            except Exception, e:
                self.warning('Unable to add slot %s: %s' % (k, e))
        # Event forwarding ...
        self.tree.refreshTree.connect(self.refreshTree)
        self.tree.nodeFound.connect(self.nodeFound)
        self.tree.deviceSelected[str].connect(self.deviceSelected)
        self.tree.search[str].connect(self.search)
        self.tree.addAttrSelected[str].connect(self.addAttrSelected)
        self.tree.removeAttrSelected[str].connect(self.removeAttrSelected)
        self.tree.modelChanged[str].connect(self.modelChanged)
        #for signal in TaurusDevTree.__pyqtSignals__:
        #    getattr(self.tree, signal).connect(
        #        lambda args, f=self, s=signal: getattr(f, signal).emit(args))
        self.edit.connectWithTree(self)
        return


##########################################################################

def taurusDevTreeMain():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.core.util import argparse

    if False:
        app = Qt.QApplication([])
        args = sys.argv[1:]
        #args = app.get_command_line_args()
        #app = TaurusApplication(sys.argv)
        if not args:
            args = ['database']
    else:
        taurus.setLogLevel(taurus.Debug)
        parser = argparse.get_taurus_parser()
        parser.set_usage("%prog [options] devname [attrs]")
        parser.set_description(
            "Taurus Application inspired in Jive and Atk Panel")
        parser.add_option("--config", "--config-file", dest="config_file", default=None,
                          help="use the given config file for initialization")
        app = TaurusApplication(cmd_line_parser=parser, app_name="TaurusDevicePanel",
                                app_version=taurus.Release.version)
        args = app.get_command_line_args()
        options = app.get_command_line_options()

    form = TaurusSearchTree()
    # try:
    # if options.tango_host is None:
    #options.tango_host = taurus.Authority().getNormalName()
    # form.setTangoHost(options.tango_host)
    # except: pass

    def trace(m):
        print(m)
    [setattr(form.tree, f, trace)
     for f in ('info', 'warning', 'error', 'trace')]

    form.setLogLevel(taurus.Debug)
    form.tree.setLogLevel(taurus.Debug)
    # set a model list from the command line or launch the chooser
    if options.config_file is not None:
        form.tree.loadConfigFile(options.config_file)
    if len(args) > 0:
        models = args
        form.setModel(models)
    form.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    taurusDevTreeMain()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a decorator to simplify the use of property."""

__all__ = ["propertx"]

__docformat__ = "restructuredtext"


def propertx(fct):
    '''
        Decorator to simplify the use of property.
        Like @property for attrs who need more than a getter.
        For getter only property use @property.

        adapted from http://code.activestate.com/recipes/502243/
    '''
    arg = [None, None, None, None]
    for i, f in enumerate(fct()):
        arg[i] = f
    if not arg[3]:
        arg[3] = fct.__doc__
    return property(*arg)

if __name__ == '__main__':

    from log import Logger

    class example(object, Logger):

        def __init__(self):
            Logger.__init__(self, "example")
            self._a = 100
            self.bar = "why"

        @propertx
        def bar():
            # BAR doc
            def get(self):
                print "\tgetting", self._a
                return self._a

            def set(self, val):
                print "\tsetting", val
                self._a = val
            return get, set

    foo = example()
    print foo.bar
    # foo.bar='egg'
    # print foo.bar



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Extension of :mod:`guiqwt.builder`"""

__all__ = ["TaurusPlotItemBuilder", "make"]

__docformat__ = 'restructuredtext'

import guiqwt.builder

from curve import TaurusCurveItem, TaurusTrendItem
from image import (TaurusImageItem, TaurusRGBImageItem, TaurusEncodedImageItem,
                   TaurusEncodedRGBImageItem, TaurusXYImageItem)
from guiqwt.curve import CurveParam
from guiqwt.image import ImageParam, XYImageItem
from guiqwt.styles import XYImageParam
from guiqwt.config import _
from guiqwt.histogram import lut_range_threshold
import numpy
from taurus.external.pint import Quantity

class TaurusPlotItemBuilder(guiqwt.builder.PlotItemBuilder):
    '''extension of :class:`guiqwt.builder.PlotItemBuilder` to provide tauruscurve and taurusimage items'''

    def __init__(self, *args, **kwargs):
        guiqwt.builder.PlotItemBuilder.__init__(self, *args, **kwargs)

    def set_curve_axes(self, *args, **kwargs):
        # ugly hack: I need to redefine this here because it is a private
        # method in PlotItemBuilder
        guiqwt.builder.PlotItemBuilder._PlotItemBuilder__set_curve_axes(
            self, *args, **kwargs)

    def set_image_param(self, *args, **kwargs):
        # ugly hack: I need to redefine this here because it is a private
        # method in PlotItemBuilder
        guiqwt.builder.PlotItemBuilder._PlotItemBuilder__set_image_param(
            self, *args, **kwargs)

    def set_param(self, *args, **kwargs):
        # ugly hack: I need to redefine this here because it is a private
        # method in PlotItemBuilder
        guiqwt.builder.PlotItemBuilder._PlotItemBuilder__set_param(
            self, *args, **kwargs)

    def pcurve(self, x, y, param, xaxis="bottom", yaxis="left"):
        """
        Extension to meth:`guiqwt.builder.PlotItemBuilder.pcurve` to support x
        and y being taurus attribute models instead of arrays, in which case, a
        TaurusCurveItem is returned. This method is called from :meth:`curve`

        Usage: pcurve(x, y, param)
        """
        if x is None or numpy.isscalar(x) or numpy.isscalar(y):
            curve = TaurusCurveItem(param)
            curve.setModels(x, y)
            curve.update_params()
            self.set_curve_axes(curve, xaxis, yaxis)
        else:
            curve = guiqwt.builder.PlotItemBuilder.pcurve(
                self, x, y, param, xaxis="bottom", yaxis="left")
        return curve

    def image(self, taurusmodel=None, **kwargs):
        """
        Extension to meth:`guiqwt.builder.PlotItemBuilder.image` to support passing a
        'taurusmodel' as a keyword argument instead passing 'data' or 'filename'.
        """
        if taurusmodel is None:
            image = guiqwt.builder.PlotItemBuilder.image(self, **kwargs)
        else:
            title = kwargs.get('title', taurusmodel)
            data = kwargs.get('data', None)
            filename = kwargs.get('filename', None)
            alpha_mask = kwargs.get('alpha_mask', None)
            alpha = kwargs.get('alpha', None)
            background_color = kwargs.get('background_color', None)
            colormap = kwargs.get('colormap', None)
            xdata = kwargs.get('xdata', [None, None])
            ydata = kwargs.get('ydata', [None, None])
            pixel_size = kwargs.get('pixel_size', None)
            interpolation = kwargs.get('interpolation', 'linear')
            eliminate_outliers = kwargs.get('eliminate_outliers', None)
            xformat = kwargs.get('xformat', '%.1f')
            yformat = kwargs.get('yformat', '%.1f')
            zformat = kwargs.get('zformat', '%.1f')
            forceRGB = kwargs.get('force_rgb', False)

            assert isinstance(xdata, (tuple, list)) and len(xdata) == 2
            assert isinstance(ydata, (tuple, list)) and len(ydata) == 2
            assert filename is None
            assert data is None

            param = ImageParam(title=_("Image"), icon='image.png')

            from taurus import Attribute
            if pixel_size is None:
                xmin, xmax = xdata
                ymin, ymax = ydata
            else:
                attr = Attribute(taurusmodel)
                valueobj = attr.read()
                data = getattr(valueobj, 'rvalue', numpy.zeros((1, 1)))
                attrdata = data
                if isinstance(data, Quantity):
                    attrdata = data.magnitude
                xmin, xmax, ymin, ymax = self.compute_bounds(attrdata,
                                                             pixel_size)

            self.set_image_param(param, title, alpha_mask, alpha, interpolation,
                                 background=background_color,
                                 colormap=colormap,
                                 xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                                 xformat=xformat, yformat=yformat,
                                 zformat=zformat)

            from taurus.core import DataType
            if forceRGB:
                if Attribute(taurusmodel).getType() == DataType.DevEncoded:
                    image = TaurusEncodedRGBImageItem(param)
                else:
                    image = TaurusRGBImageItem(param)
            else:
                if Attribute(taurusmodel).getType() == DataType.DevEncoded:
                    image = TaurusEncodedImageItem(param)
                else:
                    image = TaurusImageItem(param)
            image.setModel(taurusmodel)
            if eliminate_outliers is not None:
                image.set_lut_range(lut_range_threshold(image, 256,
                                                        eliminate_outliers))

        return image

    def xyimage(self, taurusmodel=None, **kwargs):

        if taurusmodel is None:
            return guiqwt.builder.PlotItemBuilder.xyimage(self, **kwargs)

        title = kwargs.get('title', taurusmodel)
        data = kwargs.get('data', None)
        alpha_mask = kwargs.get('alpha_mask', None)
        alpha = kwargs.get('alpha', None)
        background_color = kwargs.get('background_color', None)
        colormap = kwargs.get('colormap', None)
        interpolation = kwargs.get('interpolation', 'linear')
        eliminate_outliers = kwargs.get('eliminate_outliers', None)
        xformat = kwargs.get('xformat', '%.1f')
        yformat = kwargs.get('yformat', '%.1f')
        zformat = kwargs.get('zformat', '%.1f')
        if data is None:
            x = y = None
        else:
            x = numpy.arange(data.shape[0])
            y = numpy.arange(data.shape[1])

        data = numpy.ones((251, 251))
        x, y = numpy.arange(251) * 333, numpy.arange(251) * 1000

        param = XYImageParam(title=_("Image"), icon='image.png')
        self.set_image_param(param, title, alpha_mask, alpha, interpolation,
                             background=background_color, colormap=colormap,
                             xformat=xformat, yformat=yformat,
                             zformat=zformat)

        image = TaurusXYImageItem(param)
        image.setModel(taurusmodel)
        if eliminate_outliers is not None:
            image.set_lut_range(lut_range_threshold(image, 256,
                                                    eliminate_outliers))
        return image

    def rgbimage(self, taurusmodel=None, **kwargs):
        if taurusmodel is None:
            image = guiqwt.builder.PlotItemBuilder.rgbimage(self, **kwargs)
        else:
            image = self.image(taurusmodel=taurusmodel,
                               force_rgb=True, **kwargs)
        return image

    def ttrend(self, model, taurusparam=None, title=u"",
               color=None, linestyle=None, linewidth=None,
               marker=None, markersize=None, markerfacecolor=None,
               markeredgecolor=None, shade=None, fitted=None,
               curvestyle=None, curvetype=None, baseline=None):
        """
        Make a taurus trend item

        :return (TaurusTrendItem):
        """
        curveparam = CurveParam(icon='curve.png')
        if not title:
            title = model
        self.set_param(curveparam, title, color, linestyle, linewidth, marker,
                       markersize, markerfacecolor, markeredgecolor,
                       shade, fitted, curvestyle, curvetype, baseline)

        item = TaurusTrendItem(curveparam=curveparam, taurusparam=taurusparam)
        item.setModel(model)
        item.update_params()
        return item

#"make" is an instance of the builder (this mimics the structure of guiqwt.builder.make)
make = TaurusPlotItemBuilder()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This package contains a collection of Qt based widgets designed to interact with
taurus models. The widgets are generic in the sence that they do not assume any
behavior associated with a specific HW device. They intend to represent only
abstract model data."""

__docformat__ = 'restructuredtext'


# register icon path files and icon theme on import of taurus.qt.qtgui
import icon as __icon
import os
import glob
from taurus import tauruscustomsettings as __S

icon_dir = os.path.join(os.path.dirname(os.path.abspath(__icon.__file__)))
# TODO: get .path file glob pattern from tauruscustomsettings
__icon.registerPathFiles(glob.glob(os.path.join(icon_dir,'*.path')))
# TODO: get theme name and path from tauruscustomsettings
__icon.registerTheme(name=getattr(__S, 'QT_THEME_NAME', 'Tango'),
                     path=getattr(__S, 'QT_THEME_DIR', ''),
                     force=getattr(__S, 'QT_THEME_FORCE_ON_LINUX', False))

del os, glob, __icon, icon_dir


#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the base taurus name validator classes"""

__all__ = ["TangoAuthorityNameValidator", "TangoDeviceNameValidator",
           "TangoAttributeNameValidator"]

__docformat__ = "restructuredtext"


from taurus.core.taurusvalidator import (TaurusAttributeNameValidator,
                                         TaurusDeviceNameValidator,
                                         TaurusAuthorityNameValidator)


# todo: I do not understand the behaviour of getNames for Auth, Dev and Attr in
#      the case when the fullname does not match the regexp. For Auth it returns
#      a 3-tuple, for devs a 2-tuple and for attrs and conf a single None.
#      This is not coherent to what the method returns when it matches the
#      regexp (always a 3-tuple)

class TangoAuthorityNameValidator(TaurusAuthorityNameValidator):
    '''Validator for Tango authority names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - host: tango host name, without port.
     - port: port number
    '''

    scheme = 'tango'
    authority = '//(?P<host>([\w\-_]+\.)*[\w\-_]+):(?P<port>\d{1,5})'
    path = '(?!)'
    query = '(?!)'
    fragment = '(?!)'


class TangoDeviceNameValidator(TaurusDeviceNameValidator):
    '''Validator for Tango device names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - devname: device name (either alias or slashed name)
     - [_devalias]: device alias
     - [_devslashname]: device name in slashed (a/b/c) form
     - [host] as in :class:`TangoAuthorityNameValidator`
     - [port] as in :class:`TangoAuthorityNameValidator`

    Note: brackets on the group name indicate that this group will only contain
    a string if the URI contains it.
    '''

    scheme = 'tango'
    authority = TangoAuthorityNameValidator.authority
    path = r'/?(?P<devname>((?P<_devalias>[^/?#:]+)|' + \
           r'(?P<_devslashname>[^/?#:]+/[^/?#:]+/[^/?#:]+)))'
    query = '(?!)'
    fragment = '(?!)'

    def getNames(self, fullname, factory=None, queryAuth=True):
        '''reimplemented from :class:`TaurusDeviceNameValidator`. It accepts an
        extra keyword arg `queryAuth` which, if set to False, will prevent the
        validator from trying to query a TaurusAuthority to obtain missing info
        such as the devslashname <--> devalias correspondence.
        '''
        groups = self.getUriGroups(fullname)
        if groups is None:
            return None

        import PyTango
        default_authority = '//' + PyTango.ApiUtil.get_env_var('TANGO_HOST')

        authority = groups.get('authority')
        if authority is None:
            groups['authority'] = authority = default_authority

        db = None
        if queryAuth:
            # attempt to get an Authority object
            if factory is None:
                from taurus import Factory
                factory = Factory(scheme=self.scheme)
            try:
                db = factory.getAuthority('tango:%s' % authority)
            except:
                pass

        # note, since we validated, we either have alias or slashname (not
        # both)
        _devalias = groups.get('_devalias')
        _devslashname = groups.get('_devslashname')

        if _devslashname is None and db is not None:
            # get _devslashname from the alias using the DB
            _devslashname = db.getElementFullName(_devalias)
            groups['_devslashname'] = _devslashname

        if _devslashname is None:
            # if we still do not have a slashname, we can only give the short
            return None, None, _devalias

        # we can now construct everything. First the complete:
        complete = 'tango:%(authority)s/%(_devslashname)s' % groups

        # then the normal
        if authority.lower() == default_authority.lower():
            normal = '%(_devslashname)s' % groups
        else:
            normal = '%(authority)s/%(_devslashname)s' % groups

        # and finally the short
        if _devalias is not None:
            short = _devalias
        else:
            if db is not None:
                # get the alias from the DB (if it is defined)
                short = db.getElementAlias(_devslashname) or _devslashname
            else:
                short = _devslashname

        return complete, normal, short

    @property
    def nonStrictNamePattern(self):
        '''In non-strict mode, allow double-slash even if there is no Authority.
        (e.g., "tango://a/b/c" passes this non-strict form)
        '''
        return self.namePattern.replace('tango):(', 'tango)://(')


class TangoAttributeNameValidator(TaurusAttributeNameValidator):
    '''Validator for Tango attribute names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - attrname: attribute name including device name
     - _shortattrname: attribute name excluding device name
     - devname: as in :class:`TangoDeviceNameValidator`
     - [_devalias]: as in :class:`TangoDeviceNameValidator`
     - [_devslashname]: as in :class:`TangoDeviceNameValidator`
     - [host] as in :class:`TangoAuthorityNameValidator`
     - [port] as in :class:`TangoAuthorityNameValidator`
     - [cfgkey] same as fragment (for bck-compat use only)

    Note: brackets on the group name indicate that this group will only contain
    a string if the URI contains it.
    '''
    scheme = 'tango'
    authority = TangoAuthorityNameValidator.authority
    path = ('(?P<attrname>%s/(?P<_shortattrname>[^/?:#]+))' %
            TangoDeviceNameValidator.path)
    query = '(?!)'
    fragment = '(?P<cfgkey>[^# ]*)'

    def getNames(self, fullname, factory=None, queryAuth=True, fragment=False):
        """Returns the complete and short names"""

        groups = self.getUriGroups(fullname)
        if groups is None:
            return None

        complete, normal, short = None, None, groups.get('_shortattrname')

        # reuse the getNames from the Device validator...
        devname = fullname.rsplit('/', 1)[0]
        v = TangoDeviceNameValidator()
        devcomplete, devnormal, _ = v.getNames(devname, factory=factory,
                                               queryAuth=queryAuth)
        if devcomplete is not None:
            complete = '%s/%s' % (devcomplete, short)
        if devnormal is not None:
            normal = '%s/%s' % (devnormal, short)

        # return fragment if requested
        if fragment:
            key = groups.get('fragment', None)
            return complete, normal, short, key

        return complete, normal, short

    @property
    def nonStrictNamePattern(self):
        """In non-strict mode, allow double-slash even if there is no Authority.
        Also allow old-style "?configuration[=cfgkey]" instead of fragment.
        If cfgkey is present, it is also stored in the "fragment" named group.
        For example, "tango://a/b/c/d?configuration=label" passes this
        non-strict form, and the named group "fragment" will contain "label"
        """

        # allow for *optional* double-slashes and *optional* ?configuration...
        pattern = r'^(?P<scheme>%(scheme)s):(//)?' + \
                  r'((?P<authority>%(authority)s)(?=/))?' + \
                  r'(?P<path>%(path)s)' + \
                  r'(\?(?P<query>%(query)s))?' + \
                  r'(#%(fragment)s)?$'

        return pattern % dict(scheme=self.scheme,
                              authority='(?P<host>([\w\-_]+\.)*[\w\-_]+):(?P<port>\d{1,5})',
                              path=self.path,
                              query='configuration(=(?P<fragment>(?P<cfgkey>[^# ]+)))?',
                              fragment='(?!)')



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module sets the taurus.core.util.log.Logger to be the Qt message handler"""

__all__ = ['getQtLogger', 'initTaurusQtLogger']

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus import Logger

qtLogger = None

QT_LEVEL_MATCHER = {
    Qt.QtDebugMsg: Logger.debug,
    Qt.QtWarningMsg: Logger.warning,
    Qt.QtCriticalMsg: Logger.error,
    Qt.QtFatalMsg: Logger.error,
    Qt.QtSystemMsg: Logger.info
}


def getQtLogger():
    global qtLogger
    if qtLogger is None:
        qtLogger = Logger('QtLogger')
    return qtLogger


def qtTaurusMsgHandler(type, msg):
    global qtLogger
    if qtLogger is not None:
        caller = QT_LEVEL_MATCHER.get(type)
        caller(qtLogger, msg)


def initTaurusQtLogger():
    global qtLogger
    if not qtLogger:
        Qt.qInstallMsgHandler(qtTaurusMsgHandler)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module is designed to provide a library of taurus Qt actions"""

__all__ = ["ExternalAppAction",
           "TaurusMenu",
           "TaurusAction",
           "SeparatorAction",
           "AttributeHistoryAction",
           "AttributeAllConfigAction",
           "AttributeMonitorDeviceAction",
           "AttributeImageDisplayAction",
           "AttributeMenu",
           "ConfigurationMenu",
           ]

__docformat__ = 'restructuredtext'

import os
import xml.dom.minidom

from taurus.external.qt import Qt
from taurus.qt.qtcore.configuration import BaseConfigurableClass


class ExternalAppAction(Qt.QAction, BaseConfigurableClass):
    """ An specialized QAction for launching external applications

    Signals: apart of those from QAction, it emits a  "cmdArgsChanged" signal
    with the current cmdArgs list as its argument.
    """
    DEFAULT_ICON_NAME = 'application-x-executable'

    cmdArgsChanged = Qt.pyqtSignal(list)

    def __init__(self, cmdargs, text=None, icon=None, parent=None, interactive=True):
        '''creator

        :param cmdargs: (list<str> or str) A list of strings to be passed to
                        :meth:`subprocess.Popen` for launching the external
                        application. It can also be a single string containing a
                        command. See :meth:`setCmdArgs`
        :param text: (str) see :class:`Qt.QAction`
        :param icon: (QIcon or any other object that can be passed to QIcon creator) see :class:`Qt.QAction`
        :param parent: (QObject) The parent object
        '''
        if isinstance(cmdargs, (basestring, Qt.QString)):
            import shlex
            cmdargs = shlex.split(str(cmdargs))
        self.path = os.path.realpath(cmdargs and cmdargs[0] or '')
        if text is None:
            text = os.path.basename(cmdargs and cmdargs[0] or '')
        if icon is None:
            icon = Qt.QIcon.fromTheme(self.DEFAULT_ICON_NAME)
        elif isinstance(icon, basestring):
            tmp = Qt.QIcon.fromTheme(icon)
            if not tmp.isNull():
                icon = tmp

        Qt.QAction.__init__(self, Qt.QIcon(icon), text, parent)
        BaseConfigurableClass.__init__(self)
        self.interactive = interactive
        self._process = []
        self.setCmdArgs(cmdargs)
        self.triggered.connect(self.actionTriggered)
        self.setToolTip("Launches %s (external application)" % text)
        self.registerConfigProperty(self.cmdArgs, self.setCmdArgs, 'cmdArgs')

    def setCmdArgs(self, cmdargs, emitsignal=True):
        '''Sets the command args for executing this external application.

        It emits the "cmdArgsChanged" signal with the new cmdArgs list

        :param cmdargs: (list<str> or str) A list of strings to be passed to
                        :meth:`subprocess.Popen` for launching the external
                        application. It can also be a string containing a
                        command, which will be automatically converted to a list
        '''
        if isinstance(cmdargs, (basestring, Qt.QString)):
            import shlex
            cmdargs = shlex.split(str(cmdargs))
        self.__cmdargs = cmdargs
        if emitsignal:
            self.cmdArgsChanged.emit(self.__cmdargs)

    def cmdArgs(self):
        return self.__cmdargs

    @Qt.pyqtSlot(name='triggered', result=bool)
    def actionTriggered(self, args=None):
        '''launches the external application as a subprocess'''
        import subprocess
        try:
            if args is not None:
                if isinstance(args, (basestring, Qt.QString)):
                    import shlex
                    args = shlex.split(str(args))
                args = self.cmdArgs() + args
            else:
                args = self.cmdArgs()
            if any(args):
                #Qt.QMessageBox.warning(self.parentWidget(),'Warning','In ExternalAppAction(%s)'%args)
                self._process.append(subprocess.Popen(args))
                return True
            else:
                return False
        except OSError:
            err = "Error launching %s" % unicode(self.text())
            msg = "Cannot launch application:\n" + \
                  " ".join(self.__cmdargs) + \
                  "\nHint: Check that %s is installed and in the path" % unicode(
                      self.text())
            if self.interactive:
                Qt.QMessageBox.warning(self.parentWidget(), err, msg)
            from taurus.core.util import Logger
            Logger().warning('%s:\n%s' % (err, msg))
            return False

    def kill(self):
        # Kills all processes opened by this application
        [p.kill() for p in self._process]

    def check(self):
        '''Returns True if the application is available for executing

        :return: (bool)
        '''
        # raise NotImplementedError  #@todo: implement a checker (check if
        # self.cmdargs[0] is in the execution path and is executable
        path = os.path.realpath(self.cmdArgs()[0])
        try:
            os.stat(path)
            return True
        except:
            return False


class TaurusMenu(Qt.QMenu):
    """Base class for Taurus Menus"""

    def __init__(self, parent):
        Qt.QMenu.__init__(self, parent)

    def build(self, data):
        m_node = xml.dom.minidom.parseString(data).childNodes[0]
        self.buildFromXML(m_node)

    def getActionFactory(self):
        import taurusactionfactory
        return taurusactionfactory.ActionFactory()

    def buildFromXML(self, m_node):
        widget = self.parent()
        if m_node.hasAttribute('label'):
            self.setTitle(m_node.getAttribute('label'))

        for node in m_node.childNodes:
            name = node.nodeName
            if name == 'Menu':
                m = self.getActionFactory().buildMenu(widget, node)
                if m:
                    self.addMenu(m)
            else:
                a = self.getActionFactory().buildAction(widget, node)
                if a:
                    self.addAction(a)


class TaurusAction(Qt.QAction):
    """Base class for Taurus Actions"""

    def __init__(self, parent):
        import taurus.qt.qtgui.base

        if (parent is None) or \
                (not isinstance(parent, Qt.QWidget)) or \
                (not isinstance(parent, taurus.qt.qtgui.base.TaurusBaseComponent)):
            raise RuntimeError(
                "Invalid parent. Must be a valid Taurus widget.")

        Qt.QAction.__init__(self, parent)

        parent.modelChanged.connect(self.modelChanged)
        self.triggered.connect(self.actionTriggered)

        self.update()

    def update(self):
        model = self.parent().getModelObj()
        self.setDisabled(model is None)

    @Qt.pyqtSlot('QString')
    def modelChanged(self, modelName):
        self.update()

    @Qt.pyqtSlot(name='triggered')
    def actionTriggered(self):
        pass


class SeparatorAction(TaurusAction):

    menuID = "_Separator_"

    def __init__(self, parent=None):
        Qt.QAction.__init__(self, parent)
        self.setSeparator(True)


class AttributeHistoryAction(TaurusAction):

    menuID = "AttrHistory"

    def __init__(self, parent=None):
        TaurusAction.__init__(self, parent)
        self.setText("Show history...")

    def actionTriggered(self):
        dialog = Qt.QDialog()
        dialog.exec_()


class AttributeAllConfigAction(TaurusAction):

    menuID = "AttrConfig"

    def __init__(self, parent=None):
        TaurusAction.__init__(self, parent)
        self.setText("All...")

    def actionTriggered(self):
        taurus_widget = self.parent()
        from taurus.qt.qtgui.dialog.taurusconfigurationdialog import (
            TaurusConfigurationDialog)
        d = TaurusConfigurationDialog(taurus_widget)
        d.setModel(taurus_widget.getModelName())
        d.exec_()


class AttributeMonitorDeviceAction(TaurusAction):

    menuID = "MonitorDevice"

    def __init__(self, parent=None):
        TaurusAction.__init__(self, parent)
        self.setText("&Monitor device ...")

    def actionTriggered(self):
        taurus_widget = self.parent()
        model_name = taurus_widget.getModelName()
        w = model_name.split("/")
        if len(w) == 3:
            dev_name = model_name
        elif len(w) == 4:
            dev_name = w[0] + "/" + w[1] + "/" + w[2]
        elif len(w) == 5:  # FB: If the first parameter is the database
            dev_name = w[1] + "/" + w[2] + "/" + w[3]
        else:
            return

        cmd = "atkpanel " + dev_name + " &"
        os.system(cmd)


class AttributeImageDisplayAction(TaurusAction):

    menuID = "ImageDisplay"

    def __init__(self, parent=None):
        TaurusAction.__init__(self, parent)
        self.setText("&Image Display ...")

    def actionTriggered(self):
        taurus_widget = self.parent()
        model_name = taurus_widget.getModelName()
        w = model_name.split("/")
        if len(w) == 3:
            dev_name = model_name
        elif len(w) == 4:
            dev_name = w[0] + "/" + w[1] + "/" + w[2]
        else:
            return

        attr_name = dev_name + "/Image"

        import qub
        im = qub._TaurusQubDataImageDisplay(data=attr_name)
        im.show()


class AttributeMenu(TaurusMenu):

    menuID = "AttrMenu"
    menuData = "<Menu label='Attribute'>" \
        "<MenuItem class='AttrHistory'/>"\
        "<MenuItem class='_Separator_'/>"\
        "<Menu class='AttrConfig'/>"\
        "</Menu>"

    def __init__(self, parent):
        TaurusMenu.__init__(self, parent)
        self.build(self.menuData)


class ConfigurationMenu(TaurusMenu):
    menuID = "AttrConfigMenu"
    menuData = "<Menu label='Configuration'>" \
        "<MenuItem class='AttrConfig'/>"\
        "</Menu>"

    def __init__(self, parent):
        TaurusMenu.__init__(self, parent)
        self.build(self.menuData)



# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module exposes PyQt4/PSide uic module"""

from taurus.core.util import log
from taurus.external.qt import getQt

__backend = getQt().__name__

if __backend == 'PyQt4':
    from PyQt4.uic import *
    from PyQt4.uic import uiparser
    from PyQt4.uic import properties

    # prevent ui parser from logging debug messages
    class __IgnoreCalls:
        def __call__(self, *args, **kwargs):
            pass

    uiparser.DEBUG = __IgnoreCalls()
    properties.DEBUG = __IgnoreCalls()

    def loadUI(uiFilename, parent=None):
        newWidget = loadUi(uiFilename)
        newWidget.setParent(parent)
        return newWidget

elif __backend == 'PyQt5':
    from PyQt5.uic import *
    from PyQt5.uic import uiparser
    from PyQt5.uic import properties

    # prevent ui parser from logging debug messages
    class __IgnoreCalls:
        def __call__(self, *args, **kwargs):
            pass

    uiparser.DEBUG = __IgnoreCalls()
    properties.DEBUG = __IgnoreCalls()

    def loadUI(uiFilename, parent=None):
        newWidget = loadUi(uiFilename)
        newWidget.setParent(parent)
        return newWidget

elif __backend == 'PySide':
    from . import QtCore as __QtCore
    from . import QtUiTools as __QtUiTools

    __uiLoader = None
    class UiLoader(__QtUiTools.QUiLoader):
        def __init__(self):
            super(UiLoader, self).__init__()
            self._rootWidget = None

        def createWidget(self, className, parent=None, name=''):
            widget = super(UiLoader, self).createWidget(
                    className, parent, name)

            if name:
                if self._rootWidget is None:
                    self._rootWidget = widget
                elif not hasattr(self._rootWidget, name):
                    setattr(self._rootWidget, name, widget)
                else:
                    log.error("Qt: Name collision! Ignoring second "
                              "occurrance of %r.", name)

                if parent is not None:
                    setattr(parent, name, widget)
                else:
                    # Sadly, we can't reparent it to self, since QUiLoader
                    # isn't a QWidget.
                    log.error("Qt: No parent specified! This will probably "
                              "crash due to C++ object deletion.")

            return widget

        def load(self, fileOrName, parentWidget=None):
            if self._rootWidget is not None:
                raise Exception("UiLoader is already started loading UI!")

            widget = super(UiLoader, self).load(fileOrName, parentWidget)

            if widget != self._rootWidget:
                log.error("Qt: Returned widget isn't the root widget... ")

            self._rootWidget = None
            return widget

    def loadUI(uiFilename, parent=None):
        global __uiLoader
        if __uiLoader is None:
            __uiLoader = UiLoader()

        uiFile = __QtCore.QFile(uiFilename, parent)
        if not uiFile.open(__QtCore.QIODevice.ReadOnly):
            log.error("Qt: Couldn't open file %r!", uiFilename)
            return None

        try:
            return __uiLoader.load(uiFile, parent)

        except:
            log.exception("Exception loading UI from %r!", uiFilename)

        finally:
            uiFile.close()
            uiFile.deleteLater()

        return None

del getQt



# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
This module allows an application to provide help through the Qt
assistant tool.
The :func:`Assistant` will create a subprocess displaying the
help system for the given QtHelp collection file (.qhc).
Example usage::

    from taurus.external.qt import Qt
    from taurus.qt.qtgui.help import Assistant

    app = Qt.QApplication([])
    qas = Assistant("my_app_help.qhc")
    qas.start()
    app.exec_()
"""

__all__ = ["Assistant", "Widgets"]


from taurus.external.qt import Qt


class Widgets:
    contents = "contents"
    index = "index"
    bookmarks = "bookmarks"
    search = "search"


class _Assistant(Qt.QProcess):
    """The help assistant class"""

    def __init__(self, collection_file, parent=None):
        Qt.QProcess.__init__(self, parent)
        self.__collection_file = collection_file

    def start(self):
        if self.isRunning():
            return
        args = ["-enableRemoteControl",
                "-collectionFile", self.__collection_file]
        Qt.QProcess.start(self, "assistant", args)

    def isRunning(self):
        return self.state() == Qt.QProcess.Running

    def __send(self, cmd):
        if not self.isRunning():
            raise Exception("Assistant is not running")
        self.write(cmd + "\n")

    def assistantShow(self, widget):
        self.__send("show " + widget)

    def assistantHide(self, widget):
        self.__send("hide " + widget)

    def assistantSetSource(self, url):
        self.__send("setSource " + url)

    def assistantActivateKeyword(self, keyword):
        self.__send("activateKeyword " + keyword)

    def assistantActivateIdentifier(self, id):
        self.__send("activateIdentifier " + id)

    def assistantSyncContents(self):
        self.__send("syncContents")

    def assistantSetCurrentFilter(self, filter):
        self.__send("setCurrentFilter " + filter)

    def assistantExpandToc(self, depth):
        self.__send("expandToc " + str(depth))

    def assistantRegister(self, help_file):
        self.__send("register " + help_file)

    def assistantUnregister(self, help_file):
        self.__send("unregister " + help_file)


__ASSISTANTS = {}


def Assistant(collection_file, auto_create=True, parent=None):
    """
    The :func:`Assistant` will create a subprocess displaying the
    help system for the given QtHelp collection file (.qhc).
    Example usage::

        from taurus.external.qt import Qt
        from taurus.qt.qtgui.help import Assistant

        app = Qt.QApplication([])
        qas = Assistant("my_app_help.qhc")
        qas.start()
        app.exec_()
    """

    global __ASSISTANTS
    assistant = __ASSISTANTS.get(collection_file)
    if not auto_create:
        return assistant
    if assistant is None:
        def finished(*args):
            if __ASSISTANTS and collection_file in __ASSISTANTS:
                del __ASSISTANTS[collection_file]
        assistant = _Assistant(collection_file, parent=parent)
        __ASSISTANTS[collection_file] = assistant
        assistant.finished.connect(finished)
    return assistant


def main():
    import sys

    app = Qt.QApplication([])
    window = Qt.QWidget()
    layout = Qt.QHBoxLayout(window)
    goButton = Qt.QPushButton("Activate help", window)
    terminateButton = Qt.QPushButton("Close help", window)
    textEdit = Qt.QLineEdit(window)
    layout.addWidget(textEdit)
    layout.addWidget(goButton)
    layout.addWidget(terminateButton)

    def go():
        assistant = Assistant(textEdit.text(), parent=window)
        assistant.start()
        assistant.waitForStarted()
        assistant.assistantShow(Widgets.bookmarks)

    def terminate():
        assistant = Assistant(
            textEdit.text(), auto_create=False, parent=window)
        if assistant:
            assistant.terminate()

    goButton.clicked.connect(go)
    terminateButton.clicked.connect(terminate)
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""a list of helper methods"""

__all__ = ['check_dependencies', 'log_dependencies', 'getSchemeFromName',
           'getValidTypesForName', 'isValidName', 'makeSchemeExplicit',
           'Manager', 'Factory', 'Device', 'Attribute', 'Configuration',
           'Database', 'Authority', 'Object', 'Logger',
           'Critical', 'Error', 'Warning', 'Info', 'Debug', 'Trace',
           'setLogLevel', 'setLogFormat', 'getLogLevel', 'getLogFormat',
           'resetLogLevel', 'resetLogFormat',
           'enableLogOutput', 'disableLogOutput',
           'log', 'trace', 'debug', 'info', 'warning', 'error', 'fatal',
           'critical', 'changeDefaultPollingPeriod']

__docformat__ = "restructuredtext"

import sys
import re
from taurus import tauruscustomsettings
from .util.log import taurus4_deprecation


# regexp for finding the scheme
__SCHEME_RE = re.compile(r'([^:/?#]+):.*')


def __translate_version_str2int(version_str):
    """Translates a version string in format x[.y[.z[...]]] into a 000000 number"""
    import math
    parts = version_str.split('.')
    i, v, l = 0, 0, len(parts)
    if not l:
        return v
    while i < 3:
        try:
            v += int(parts[i]) * int(math.pow(10, (2 - i) * 2))
            l -= 1
            i += 1
        except ValueError, ve:
            return v
        if not l:
            return v
    return v

    try:
        v += 10000 * int(parts[0])
        l -= 1
    except ValueError, ve:
        return v
    if not l:
        return v

    try:
        v += 100 * int(parts[1])
        l -= 1
    except ValueError, ve:
        return v
    if not l:
        return v

    try:
        v += int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l:
        return v


def __get_python_version():
    return '.'.join(map(str, sys.version_info[:3]))


def __get_python_version_number():
    pyver_str = __get_python_version()
    if pyver_str is None:
        return None
    return __translate_version_str2int(pyver_str)


def __get_pytango_version():
    try:
        import PyTango
    except:
        return None
    try:
        return PyTango.Release.version
    except:
        return '0.0.0'


def __get_pytango_version_number():
    tgver_str = __get_pytango_version()
    if tgver_str is None:
        return None
    return __translate_version_str2int(tgver_str)


def __get_pyqt_version():
    try:
        import PyQt4.Qt
        return PyQt4.Qt.PYQT_VERSION_STR
    except:
        return None


def __get_pyqt_version_number():
    pyqtver_str = __get_pyqt_version()
    if pyqtver_str is None:
        return None
    return __translate_version_str2int(pyqtver_str)


def __get_pyqwt_version():
    try:
        import PyQt4.Qwt5
        return PyQt4.Qwt5.QWT_VERSION_STR
    except:
        pass


def __get_pyqwt_version_number():
    pyqwtver_str = __get_pyqwt_version()
    if pyqwtver_str is None:
        return None
    return __translate_version_str2int(pyqwtver_str)


def __get_qub_version():
    try:
        import Qub4
        return "1.0.0"
    except:
        try:
            import Qub
            return "1.0.0"
        except:
            pass


def __get_qub_version_number():
    qubver_str = __get_qub_version()
    if qubver_str is None:
        return None
    return __translate_version_str2int(qubver_str)


def __get_qtcontrols_version():
    try:
        import qtcontrols
        return "1.0.0"
    except:
        pass


def __get_qtcontrols_version_number():
    qtcontrols_str = __get_qtcontrols_version()
    if qtcontrols_str is None:
        return None
    return __translate_version_str2int(qtcontrols_str)


def __get_spyderlib_version():
    try:
        import spyderlib
        return spyderlib.__version__
    except:
        pass


def __get_spyderlib_version_number():
    spyderlibver_str = __get_spyderlib_version()
    if spyderlibver_str is None:
        return None
    return __translate_version_str2int(spyderlibver_str)


def __w(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()


def __wn(msg):
    __w(msg + '\n')


def check_dependencies():
    for msg in _check_dependencies(forlog=False):
        m = msg[1]
        if msg[0] != -1:
            m = '\t%s' % m
        print m


def log_dependencies():
    from taurus.core.util.log import Logger
    l = Logger("taurus")
    for msg in _check_dependencies(forlog=True):
        if msg[0] != -1:
            l.info(msg[1])


def _check_dependencies(forlog=False):
    """Checks for the required and optional packages of taurus"""

    if forlog:
        MSG = {'OK': '[OK]', 'ERR': '[ERROR]', 'WARN': '[WARNING]'}
    else:
        MSG = {
            'OK': "[\033[0;32mOK\033[0m]",
            'ERR': "[\033[0;31mERROR\033[0m]",
            'WARN': "[\033[0;33mWARNING\033[0m]"}

    core_requirements = {
        #    module       minimum  recommended
        "Python": ("2.6.0", "2.6.0"),
    }

    core_optional_requirements = {
        #    module       minimum  recommended
        "PyTango": ("7.1.0", "7.1.0"),
    }

    widget_requirements = {
        #    module       minimum  recommended
        "PyTango": ("7.1.0", "7.1.0"),
        "PyQt": ("4.4.0", "4.4.0"),
        "PyQwt": ("5.2.0", "5.2.0"),
    }

    widget_optional_requirements = {
        #    module       minimum  recommended
        "Qub": ("1.0.0", "1.0.0"),
        "qtcontrols": ("1.0.0", "1.0.0"),
        "spyderlib": ("2.0.0", "2.0.0"),
    }

    yield -1, "Checking required dependencies of taurus.core..."
    r = core_requirements

    m = "Checking for Python >=%s..." % r["Python"][0]
    minPython, recPython = map(__translate_version_str2int, r["Python"])
    currPython, currPythonStr = __get_python_version_number(), __get_python_version()
    if currPython is None:
        yield 2, "{msg} {ERR} (Not found])".format(msg=m, **MSG)
    elif currPython < minPython:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currPythonStr, rec=r['Python'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currPythonStr, **MSG)

    yield -1, "Checking OPTIONAL dependencies of taurus.core..."
    r = core_optional_requirements

    m = "Checking for PyTango >=%s..." % r["PyTango"][0]
    minPyTango, recPyTango = map(__translate_version_str2int, r["PyTango"])
    currPyTango, currPyTangoStr = __get_pytango_version_number(), __get_pytango_version()
    if currPyTango is None:
        yield 2, "{msg} {ERR} (Not found])".format(msg=m, **MSG)
    elif currPyTango < minPyTango:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currPyTangoStr, rec=r['PyTango'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currPyTangoStr, **MSG)

    yield -1, "Checking required dependencies of taurus.qt..."
    r = widget_requirements

    m = "Checking for PyTango >=%s..." % r["PyTango"][0]
    if currPyTango is None:
        yield 2, "{msg} {ERR} (Not found])".format(msg=m, **MSG)
    elif currPyTango < minPyTango:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currPyTangoStr, rec=r['PyTango'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currPyTangoStr, **MSG)

    m = "Checking for PyQt >=%s..." % r["PyQt"][0]
    minPyQt, recPyQt = map(__translate_version_str2int, r["PyQt"])
    currPyQt, currPyQtStr = __get_pyqt_version_number(), __get_pyqt_version()
    if currPyQt is None:
        yield 2, "{msg} {ERR} (Not found])".format(msg=m, **MSG)
    elif currPyQt < minPyQt:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currPyQtStr, rec=r['PyQt'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currPyQtStr, **MSG)

    m = "Checking for PyQwt >=%s..." % r["PyQwt"][0]
    minPyQwt, recPyQwt = map(__translate_version_str2int, r["PyQwt"])
    currPyQwt, currPyQwtStr = __get_pyqwt_version_number(), __get_pyqwt_version()
    if currPyQwt is None:
        yield 1, "{msg} {ERR} (Not found])".format(msg=m, **MSG)
    elif currPyQwt < minPyQwt:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currPyQwtStr, rec=r['PyQwt'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currPyQwtStr, **MSG)

    yield -1, "Checking OPTIONAL dependencies of taurus.qt..."
    r = widget_optional_requirements

    m = "Checking for Qub >=%s..." % r["Qub"][0]
    minQub, recQub = map(__translate_version_str2int, r["Qub"])
    currQub, currQubStr = __get_qub_version_number(), __get_qub_version()
    if currQub is None:
        yield 1, "{msg} {WARN} (Not found])".format(msg=m, **MSG)
    elif currQub < minQub:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currQubStr, rec=r['Qub'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currQubStr, **MSG)

    m = "Checking for spyderlib >=%s..." % r["spyderlib"][0]
    minspyderlib, recspyderlib = map(
        __translate_version_str2int, r["spyderlib"])
    currspyderlib, currspyderlibStr = __get_spyderlib_version_number(
    ), __get_spyderlib_version()
    if currspyderlib is None:
        yield 1, "{msg} {WARN} (Not found])".format(msg=m, **MSG)
    elif currspyderlib < minspyderlib:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currspyderlibStr, rec=r['spyderlib'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currspyderlibStr, **MSG)

    m = "Checking for qtcontrols >=%s..." % r["qtcontrols"][0]
    minqtcontrols, recqtcontrols = map(
        __translate_version_str2int, r["qtcontrols"])
    currqtcontrols, currqtcontrolsStr = __get_qtcontrols_version_number(
    ), __get_qtcontrols_version()
    if currqtcontrols is None:
        yield 1, "{msg} {WARN} (Not found])".format(msg=m, **MSG)
    elif currqtcontrols < minqtcontrols:
        yield 1, "{msg} {WARN} (Found {fnd}. Recommended >={rec})".format(msg=m, fnd=currqtcontrolsStr, rec=r['qtcontrols'][1], **MSG)
    else:
        yield 0, "{msg} {OK} (Found {fnd})".format(msg=m, fnd=currqtcontrolsStr, **MSG)


def getSchemeFromName(name, implicit=True):
    """Return the scheme from a taurus name.

    :param name: (str) taurus model name URI.
    :param implicit: (bool) controls whether to return the default scheme
                     (if implicit is True -default-) or None (if implicit is
                     False) in case `model` does not contain the scheme name
                     explicitly. The default scheme may be defined in
                     :ref:`tauruscustomsettings` ('tango' is assumed if
                     not defined)
    """
    m = __SCHEME_RE.match(name)
    if m is not None:
        return m.groups()[0]
    if implicit:
        return getattr(tauruscustomsettings, 'DEFAULT_SCHEME', "tango")
    else:
        return None


def makeSchemeExplicit(name, default=None):
    """return the name guaranteeing that the scheme is present. If name already
    contains the scheme, it is returned unchanged.

    :param name: (str) taurus model name URI.
    :param default: (str) The default scheme to use. If no default is passed,
                     the one defined in tauruscustomsettings.DEFAULT_SCHEME is
                     used.

    :return: the name with the explicit scheme.
    """
    if getSchemeFromName(name, implicit=False) is None:
        if default is None:
            default = getattr(tauruscustomsettings, 'DEFAULT_SCHEME', "tango")
        return "%s:%s" % (default, name)
    else:
        return name


def getValidTypesForName(name, strict=None):
    """
    Returns a list of all Taurus element types for which `name` is a valid
    model name (while in many cases a name may only be valid for one
    element type, this is not necessarily true in general)

    :param name: (str) taurus model name
    :param strict: (bool) If True, names that are not RFC3986-compliant but
                   which would be accepted for backwards compatibility are
                   considered valid.

    :return: (list<TaurusElementType.element>) where element can be one of:
             `Attribute`, `Device` or `Authority`
    """
    try:
        factory = Factory(scheme=getSchemeFromName(name))
    except:
        return []
    return factory.getValidTypesForName(name, strict=strict)


def isValidName(name, etypes=None, strict=None):
    """Returns True is the given name is a valid Taurus model name. If
    `etypes` is passed, it returns True only if name is valid for at least
    one of the given the element types. Otherwise it returns False.
    For example::

        isValidName('tango:foo')--> True
        isValidName('tango:a/b/c', [TaurusElementType.Attribute]) --> False

    :param name: (str) the string to be checked for validity
    :param etypes: (seq<TaurusElementType>) if given, names will only be
                   considered valid if they represent one of the given
                   element types. Supported element types are:
                   `Attribute`, `Device` and `Authority`
    :param strict: (bool) If True, names that are not RFC3986-compliant but
                   which would be accepted for backwards compatibility are
                   considered valid.

    :return: (bool)
    """
    validtypes = getValidTypesForName(name, strict=strict)
    if etypes is None:
        return bool(validtypes)
    for e in etypes:
        if e in validtypes:
            return True
        return False


def Manager():
    """Returns the one and only TaurusManager

    It is a shortcut to::

        import taurus.core
        manager = taurus.core.taurusmanager.TaurusManager()

    :return: the TaurusManager
    :rtype: :class:`taurus.core.taurusmanager.TaurusManager`

    .. seealso:: :class:`taurus.core.taurusmanager.TaurusManager`
    """
    from taurus.core.taurusmanager import TaurusManager
    return TaurusManager()


def Factory(scheme=None):
    """Returns the one and only Factory for the given scheme

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory(scheme)

    :param scheme: a string representing the scheme. Default value is None meaning ``tango`` scheme
    :type scheme: str
    :return: a taurus factory
    :rtype: :class:`taurus.core.taurusfactory.TaurusFactory`
    """
    manager = Manager()
    f = manager.getFactory(scheme=scheme)
    if f is None:
        from taurus.core.taurusexception import TaurusException
        if scheme is None:
            scheme = "default scheme '" + manager.default_scheme + "'"
        else:
            scheme = "'" + scheme + "'"
        raise TaurusException('Cannot create Factory for %s' % scheme)
    return f()


def Device(device_name):
    """Returns the taurus device for the given device name

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        device  = factory.getDevice(device_name)

    :param device_name: the device name
    :type device_name: str
    :return: a taurus device
    :rtype: :class:`taurus.core.taurusdevice.TaurusDevice`
    """
    return Factory(scheme=getSchemeFromName(device_name)).getDevice(device_name)


def Attribute(dev_or_attr_name, attr_name=None):
    """Returns the taurus attribute for either the pair (device name, attribute name)
    or full attribute name

    - Attribute(full_attribute_name)
    - Attribute(device_name, attribute_name)

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        attribute  = factory.getAttribute(full_attribute_name)

    or::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        device  = factory.getDevice(device_name)
        attribute = device.getAttribute(attribute_name)

    :param dev_or_attr_name: the device name or full attribute name
    :type dev_or_attr_name: str or TaurusDevice
    :param attr_name: attribute name
    :type attr_name: str
    :return: a taurus attribute
    :rtype: :class:`taurus.core.taurusattribute.TaurusAttribute`
    """
    import types

    if attr_name is None:
        return Factory(scheme=getSchemeFromName(dev_or_attr_name)).getAttribute(dev_or_attr_name)
    else:
        if type(dev_or_attr_name) in types.StringTypes:
            dev = Device(dev_or_attr_name)
        else:
            dev = dev_or_attr_name
        return dev.getAttribute(attr_name)


@taurus4_deprecation(alt='Attribute')
def Configuration(attr_or_conf_name, conf_name=None):
    """Returns the taurus configuration for either the pair
    (attribute name, conf name) or full conf name

    - Configuration(full_conf_name)
    - Configuration(attribute_name, conf_name)

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        conf  = factory.getConfiguration(attr_or_conf_name)

    or::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        attribute  = factory.getAttribute(attribute_name)
        conf = attribute.getConfig(conf_name)

    :param attr_or_conf_name: the full attribute name or full conf name
    :type attr_or_conf_name: str
    :param conf_name: conf name
    :type conf_name: str or None
    :return: a taurus configuration
    :rtype: :class:`taurus.core.taurusconfiguration.TaurusConfiguration`
    """
    return Attribute(attr_or_conf_name)


@taurus4_deprecation(alt='Authority')
def Database(name=None):
    return Authority(name=name)


def Authority(name=None):
    """Returns a taurus authority

    It is a shortcut to::

        import taurus.core.taurusmanager
        manager = taurus.core.taurusmanager.TaurusManager()
        factory = manager.getFactory()
        db  = factory.getAuthority(dname)

    :param name: authority name. If None (default) it will return the default
                 authority of the default scheme. For example, if the default
                 scheme is tango, it will return the default TANGO_HOST database
    :type name: str or None
    :return: a taurus authority
    :rtype: :class:`taurus.core.taurusauthority.TaurusAuthority`
    """
    return Factory(getSchemeFromName(name or '')).getAuthority(name)


def Object(*args):
    """Returns an taurus object of given class for the given name

    Can be called as:

      - Object(name)
      - Object(cls, name)

    Where:

      - `name` is a model name (str)
      - `cls` is a class derived from TaurusModel

    If `cls` is not given, Object() will try to guess it from `name`.

    :return: a taurus object
    :rtype: :class:`taurus.core.taurusmodel.TaurusModel`
    """
    if len(args) == 1:
        klass, name = None, args[0]
    elif len(args) == 2:
        klass, name = args
    else:
        msg = 'Object() takes either 1 or 2 arguments (%i given)' % len(args)
        raise TypeError(msg)
    factory = Factory(getSchemeFromName(name))
    if klass is None:
        klass = factory.findObjectClass(name)
    return factory.getObject(klass, name)

from taurus.core.util import log as __log_mod

Logger = __log_mod.Logger
Critical = Logger.Critical
Fatal = Logger.Fatal
Error = Logger.Error
Warning = Logger.Warning
Info = Logger.Info
Debug = Logger.Debug
Trace = Logger.Trace

setLogLevel = Logger.setLogLevel
setLogFormat = Logger.setLogFormat
getLogLevel = Logger.getLogLevel
getLogFormat = Logger.getLogFormat
resetLogLevel = Logger.resetLogLevel
resetLogFormat = Logger.resetLogFormat

enableLogOutput = Logger.enableLogOutput
disableLogOutput = Logger.disableLogOutput

log = __log_mod._log
trace = __log_mod.trace
debug = __log_mod.debug
info = __log_mod.info
warning = __log_mod.warning
error = __log_mod.error
fatal = __log_mod.fatal
critical = __log_mod.critical
deprecated = __log_mod.deprecated


def changeDefaultPollingPeriod(period):
    Manager().changeDefaultPollingPeriod(period)

#del __log_mod
#del __translate_version_str2int



#!/usr/bin/env python

##############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import os
import imp
from setuptools import setup, find_packages


def get_release_info():
    name = "release"
    setup_dir = os.path.dirname(os.path.abspath(__file__))
    release_dir = os.path.join(setup_dir, 'lib', 'taurus', 'core')
    data = imp.find_module(name, [release_dir])
    ret = imp.load_module(name, *data)
    return ret

release = get_release_info()

package_dir = {'': 'lib'}

packages = find_packages(where='lib')

provides = [
    'taurus',
    'taurus.core',
    'taurus.qt',
    # 'Taurus-Tango',  # [Taurus-Tango]
    # 'Taurus-Qt',  # [Taurus-Qt]
    # 'Taurus-Qt-PyQwt',  # [Taurus-Qt-Plot]
    # 'Taurus-Qt-Synoptic',  # [Taurus-Qt-Synoptic]
    # 'Taurus-TaurusGUI',  # [Taurus-TaurusGUI]
    # 'Taurus-Editor',  # [Taurus-Editor] --> or maybe move it to sardana
    # 'Taurus-Qt-Guiqwt',
]

requires = [
    'numpy (>=1.1)',
    'PyTango (>=7.1)',  # [Taurus-Tango]
    'PyQt4 (>=4.8)',  # [Taurus-Qt]
    'PyQt4.Qwt5 (>=5.2.0)',  # [Taurus-Qt-Plot]
    'ply (>=2.3)',  # [Taurus-Qt-Synoptic]
    'lxml (>=2.1)',  # [Taurus-TaurusGUI]
    'spyder (>=2.2)',  # [Taurus-Editor] --> or maybe move it to sardana
    'guiqwt (==2.3.1)',
]

extras_require = {
    # 'Taurus-Tango': ['PyTango>=7.1'],  # [Taurus-Tango]
    # 'Taurus-Qt': ['PyQt4>=4.8'],  # [Taurus-Qt]
    # 'Taurus-Qt-PyQwt': ['PyQt4.Qwt5>=5.2.0'],  # [Taurus-Qt-Plot]
    # 'Taurus-Qt-Synoptic': ['ply>=2.3'],  # [Taurus-Qt-Synoptic]
    # 'Taurus-TaurusGUI': ['lxml>=2.1'],  # [Taurus-TaurusGUI]
    # 'Taurus-Editor': ['spyder>=2.2'],  # [Taurus-Editor] --> or maybe move it to sardana
    # 'Taurus-Qt-Guiqwt': ['guiqwt==2.3.1','guidata==1.6.1'],
    # 'Taurus-Nexus': ['PyMca<5'],
}

console_scripts = [
    'taurustestsuite = taurus.test.testsuite:main',
    # TODO: taurusdoc,
]

gui_scripts = [
    'taurusconfigbrowser = taurus.qt.qtgui.panel.taurusconfigeditor:main',
    'taurusplot = taurus.qt.qtgui.plot.taurusplot:main',
    'taurustrend = taurus.qt.qtgui.plot.taurustrend:main',
    'taurusform = taurus.qt.qtgui.panel.taurusform:taurusFormMain',
    'tauruspanel = taurus.qt.qtgui.panel.taurusdevicepanel:TaurusPanelMain',
    'taurusdevicepanel = taurus.qt.qtgui.panel.taurusdevicepanel:TaurusDevicePanelMain',
    'taurusgui = taurus.qt.qtgui.taurusgui.taurusgui:main',
    'taurusdesigner = taurus.qt.qtdesigner.taurusdesigner:main',
    'tauruscurve = taurus.qt.qtgui.extra_guiqwt.plot:taurusCurveDlgMain',
    'taurustrend1d = taurus.qt.qtgui.extra_guiqwt.plot:taurusTrendDlgMain',
    'taurusimage = taurus.qt.qtgui.extra_guiqwt.plot:taurusImageDlgMain',
    'taurustrend2d = taurus.qt.qtgui.extra_guiqwt.taurustrend2d:taurusTrend2DMain',
    'taurusiconcatalog = taurus.qt.qtgui.icon.catalog:main',
    # 'taurusconfigbrowser = taurus.qt.qtgui.panel.taurusconfigeditor:main [Taurus-Qt]',
    # 'taurusplot = taurus.qt.qtgui.plot.taurusplot:main [Taurus-Qt-Plot]',
    # 'taurustrend = taurus.qt.qtgui.plot.taurustrend:main [Taurus-Qt-Plot]',
    # 'taurusform = taurus.qt.qtgui.panel.taurusform:taurusFormMain [Taurus-Qt]',
    # 'tauruspanel = taurus.qt.qtgui.panel.taurusdevicepanel:TaurusPanelMain [Taurus-Qt]',
    # 'taurusdevicepanel = taurus.qt.qtgui.panel.taurusdevicepanel:TaurusDevicePanelMain [Taurus-Qt-Plot]',
    # 'taurusgui = taurus.qt.qtgui.taurusgui.taurusgui:main [Taurus-Qt-Plot,Taurus-Qt-Synoptic]',
    # 'taurusdesigner = taurus.qt.qtdesigner.taurusdesigner:main [Taurus-Qt]',
    # 'tauruscurve = taurus.qt.qtgui.extra_guiqwt.plot:taurusCurveDlgMain [Taurus-Qt-Guiqwt]',
    # 'taurustrend1d = taurus.qt.qtgui.extra_guiqwt.plot:taurusTrendDlgMain [Taurus-Qt-Guiqwt]',
    # 'taurusimage = taurus.qt.qtgui.extra_guiqwt.plot:taurusImageDlgMain [Taurus-Qt-Guiqwt]',
    # 'taurustrend2d = taurus.qt.qtgui.extra_guiqwt.taurustrend2d:taurusTrend2DMain [Taurus-Qt-Guiqwt]',
    # TODO: taurusremotelogmonitor
    # TODO: taurusdemo
]

entry_points={'console_scripts': console_scripts,
              'gui_scripts': gui_scripts,
              }

classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Win32 (MS Windows)',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: User Interfaces',
    'Topic :: Software Development :: Widget Sets',
]


setup(name='taurus',
      version=release.version,
      description=release.description,
      long_description=release.long_description,
      author=release.authors['Tiago_et_al'][0],
      maintainer=release.authors['Community'][0],
      maintainer_email=release.authors['Community'][1],
      url=release.url,
      download_url=release.download_url,
      platforms=release.platforms,
      license=release.license,
      keywords=release.keywords,
      packages=packages,
      package_dir=package_dir,
      classifiers=classifiers,
      include_package_data=True,
      entry_points=entry_points,
      provides=provides,
      requires=requires,
      extras_require=extras_require,
      test_suite='taurus.test.testsuite.get_suite',
      )
